"""Provide a decorator to register event handlers."""


def socketio_handler(event, namespace=None):
    """Register a socketio handler via decorator."""

    def wrapper(func):
        """Decorate a ws event handler."""
        # pylint: disable=protected-access
        func._ws_event = event
        func._ws_namespace = namespace
        return func

    return wrapper


def partial_coro(func, *args, **keywords):
    """Return a partial coroutine."""
    # https://docs.python.org/3/library/functools.html#functools.partial
    async def wrapper(*fargs, **fkeywords):
        """Wrap the coroutine function."""
        newkeywords = keywords.copy()
        newkeywords.update(fkeywords)
        return await func(*args, *fargs, **newkeywords)

    wrapper.func = func
    wrapper.args = args
    wrapper.keywords = keywords
    return wrapper



"""Provide worker functions for Python 3."""
import asyncio
import logging
import sys
import traceback

import janus

from .utils import format_traceback
from .python3_client import AsyncClient
from .python_worker import PluginConnection
from ipykernel.comm import Comm
from IPython.display import HTML
from imjoy.utils import dotdict

import os

# pylint: disable=unused-argument, redefined-outer-name

logger = logging.getLogger("jupyter_client")


def show_imjoy(client_id, url="https://imjoy.io/#/app", width="100%", height=650):
    """Show ImJoy in the output cell."""
    iframe_html = """
    <style>
    .card {
        background: #fff;
        border-radius: 2px;
        display: inline-block;
        height: 550px;
        margin: 1rem;
        position: relative;
        width: calc(100%% - 3ex);
        border-radius: 4px;
    }
    .card-1 {
        box-shadow: 0 1px 2px rgba(0,0,0,0.12), 0 1px 2px rgba(0,0,0,0.24);
        transition: all 0.3s cubic-bezier(.25,.8,.25,1);
    }
    .card-1:hover {
        box-shadow: 0 6px 10px rgba(0,0,0,0.25), 0 3px 3px rgba(0,0,0,0.22);
    }
    </style>
    <iframe id="iframe_%(client_id)s" onload="setup_imjoy_bridge()" src="%(url)s" class="card card-1" frameborder="0" width="%(width)s" height="%(height)s" style="max-width: 100%%;"></iframe>
    <script type="text/Javascript">
    (function(){
        class PostMessageIO{
            constructor(iframeEl){
                this.plugins = {}
                this._callbacks = {}
                this.iframeEl = iframeEl
                window.addEventListener(
                    "message",
                    (e)=>{
                        if(e.data.pio_connect){
                            if(this._callbacks['connect']){
                                for(let cb of this._callbacks['connect']){ 
                                    try{
                                        cb()
                                    }
                                    catch(e){
                                        console.error(e)
                                    }
                                }
                            }
                        }
                        else if(e.data.pio_disconnect){
                            if(this._callbacks['disconnect']){
                                for(let cb of this._callbacks['disconnect']){ 
                                    try{
                                        cb()
                                    }
                                    catch(e){
                                        console.error(e)
                                    }
                                }
                            }
                        }
                        else{
                            if (iframeEl && iframeEl.contentWindow && e.source === iframeEl.contentWindow) {
                                if(e.data.channel && this._callbacks[e.data.channel]){
                                    for(let cb of this._callbacks[e.data.channel]){
                                        try{
                                            cb(e.data)
                                        }
                                        catch(e){
                                            console.error(e)
                                        }
                                    }
                                }
                            }
                        }
                    },
                    false
                );
            }
            
            emit(channel, data, transferables) {
                data.channel = channel
                this.iframeEl.contentWindow.postMessage(
                    data,
                    "*",
                    transferables
                );
            }

            on(channel, callback){
                if(this._callbacks[channel]){
                    this._callbacks[channel].push(callback)
                }
                else{
                    this._callbacks[channel] = [callback]
                }
            }
            
            off(channel, callback){
                if(this._callbacks[channel]){
                    for( var i = 0; i < this._callbacks[channel].length; i++){ 
                        if ( this._callbacks[channel][i] === callback) {
                            this._callbacks[channel].splice(i, 1); 
                        }
                    }
                }
            }
        }
        window.PostMessageIO = PostMessageIO;
    })()

    function setup_imjoy_bridge(){
        // try to recover the client
        try{
            var kernel = IPython.notebook.kernel;
            command = 'from imjoy.workers.jupyter_client import JupyterClient;JupyterClient.recover_client("%(client_id)s")';
            kernel.execute(command);
        }
        catch(e){
            console.warn('Failed to run recover client command.', e)
        }

        const iframeEl = document.getElementById('iframe_%(client_id)s')
        const pio = new PostMessageIO(iframeEl);
        let _connected_comm = null;
        function add_plugin(plugin, _connected_comm){
            const id = plugin.id
            pio.on("message_to_plugin_" + id, (data)=>{
                _connected_comm.send(data.data);
            })
            _connected_comm.on_msg((msg) => {
                var data = msg.content.data
                if (["initialized",
                    "importSuccess",
                    "importFailure",
                    "executeSuccess",
                    "executeFailure"
                    ].includes(data.type)) {
                    pio.emit("message_from_plugin_" + id, data);
                } else {
                    pio.emit("message_from_plugin_" + id, { type: 'message', data: data });
                }
                
            })
        }
        Jupyter.notebook.kernel.comm_manager.register_target(
            'imjoy_comm_%(client_id)s',
            function (comm, open_msg) {
                _connected_comm = comm;
                for(let k in pio.plugins){
                    add_plugin(pio.plugins[k], _connected_comm)
                }
                //var config = open_msg.content.data
                //pio.emit("message_from_plugin_" + id, {'type': 'init_plugin', 'id': config.id, 'config': config});
            }
        )

        pio.on('connect', ()=>{
            console.log('pio connected.')
        })

        pio.on('init_plugin', (plugin_config)=>{
            //id, type, config
            pio.plugins[plugin_config.id] = plugin_config
            if(_connected_comm){
                add_plugin(plugin_config, _connected_comm)
            }
            try{
                var kernel = IPython.notebook.kernel;
                command = 'from imjoy.workers.python_worker import PluginConnection as __plugin_connection__;__plugin_connection__.add_plugin("'+plugin_config.id+'", "%(client_id)s").start()';
                kernel.execute(command);
            }
            catch(e){
                console.error('Failed to execute command to start the plugin', plugin_config, e)
            }
        });
    }
    </script>
    """ % {
        "url": url,
        "width": width,
        "height": height,
        "client_id": client_id,
    }
    return HTML(iframe_html)


class JupyterClient(AsyncClient):
    """Represent an async socketio client."""

    @staticmethod
    def recover_client(id):
        if JupyterClient._clients.get(id):
            return JupyterClient._clients.get(id)
        else:
            return JupyterClient(id)

    # pylint: disable=too-few-public-methods
    def __init__(self, id=None):
        """Set up client instance."""
        super().__init__(id)
        self.comm = None

    def setup(self, conn):
        """Set up the plugin connection."""

        self.comm = self.comm or Comm(
            target_name="imjoy_comm_" + self.id, data={"id": self.id}
        )

        def on_disconnect():
            if not conn.opt.daemon:
                conn.exit(1)

        def emit(msg):
            """Emit a message to the socketio server."""
            self.comm.send(msg)

        def comm_plugin_message(msg):
            """Handle plugin message."""

            data = msg["content"]["data"]
            # emit({'type': 'logging', 'details': data})

            # if not self.conn.executed:
            #    self.emit({'type': 'message', 'data': {"type": "interfaceSetAsRemote"}})
            if data["type"] == "import":
                emit({"type": "importSuccess", "url": data["url"]})
            elif data["type"] == "disconnect":
                conn.abort.set()
                try:
                    if "exit" in conn.interface and callable(conn.interface["exit"]):
                        conn.interface["exit"]()
                except Exception as exc:  # pylint: disable=broad-except
                    logger.error("Error when exiting: %s", exc)
            elif data["type"] == "execute":
                if not conn.executed:
                    self.queue.put(data)
                else:
                    logger.debug("Skip execution")
                    emit({"type": "executeSuccess"})
            elif data["type"] == "message":
                _data = data["data"]
                self.queue.put(_data)
                logger.debug("Added task to the queue")

        self.comm.on_msg(comm_plugin_message)
        self.comm.on_close(on_disconnect)

        conn.default_exit = lambda: None
        conn.emit = emit

        emit({"type": "initialized", "dedicatedThread": True})
        logger.info("Plugin %s initialized", conn.opt.id)

    # def run_forever(self, conn):
    #     self.loop.create_task(self.task_worker(conn, self.queue, logger, conn.abort))

    def start(
        self,
        name="Untitled Plugin",
        workspace="default",
        imjoy_url="https://imjoy.io",
        width="100%",
        height=650,
    ):
        """Show ImJoy in the output cell."""
        return show_imjoy(
            self.id,
            url=f"{imjoy_url}/#/app?jupyter_plugin={name}&workspace={workspace}",
            width=width,
            height=height,
        )



"""Provide utils for Python 2 plugins."""
import copy
import threading
import time
import uuid
from importlib import import_module


class Registry(dict):
    """Registry of items."""

    # https://github.com/home-assistant/home-assistant/blob/
    # 2a9fd9ae269e8929084e53ab12901e96aec93e7d/homeassistant/util/decorator.py
    def register(self, name):
        """Return decorator to register item with a specific name."""

        def decorator(func):
            """Register decorated function."""
            self[name] = func
            return func

        return decorator


def get_psutil():
    """Try to import and return psutil."""
    try:
        return import_module("psutil")
    except ImportError:
        print(
            "WARNING: a library called 'psutil' can not be imported, "
            "this may cause problem when killing processes."
        )
        return None


def debounce(secs):
    """Decorate to ensure function can only be called once every `s` seconds."""

    def decorate(func):
        store = {"t": None}

        def wrapped(*args, **kwargs):
            if store["t"] is None or time.time() - store["t"] >= secs:
                result = func(*args, **kwargs)
                store["t"] = time.time()
                return result
            return None

        return wrapped

    return decorate


def set_interval(interval):
    """Set interval."""

    def decorator(function):
        def wrapper(*args, **kwargs):
            stopped = threading.Event()

            def loop():  # executed in another thread
                while not stopped.wait(interval):  # until stopped
                    function(*args, **kwargs)

            thread = threading.Thread(target=loop)
            thread.daemon = True  # stop if the program exits
            thread.start()
            return stopped

        return wrapper

    return decorator


class dotdict(dict):  # pylint: disable=invalid-name
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))


def get_key_by_value(dict_, value):
    """Return key by value."""
    for key, val in dict_.items():
        if value == val:
            return key
    return None


def format_traceback(traceback_string):
    """Format traceback."""
    formatted_lines = traceback_string.splitlines()
    # remove the second and third line
    formatted_lines.pop(1)
    formatted_lines.pop(1)
    formatted_error_string = "\n".join(formatted_lines)
    formatted_error_string = formatted_error_string.replace(
        'File "<string>"', "Plugin script"
    )
    return formatted_error_string


class ReferenceStore:
    """Represent a reference store."""

    def __init__(self):
        """Set up store."""
        self._store = {}

    @staticmethod
    def _gen_id():
        """Generate an id."""
        return str(uuid.uuid4())

    def put(self, obj):
        """Put an object into the store."""
        id_ = self._gen_id()
        self._store[id_] = obj
        return id_

    def fetch(self, search_id):
        """Fetch an object from the store by id."""
        if search_id not in self._store:
            return None
        obj = self._store[search_id]
        if not hasattr(obj, "__remote_method"):
            del self._store[search_id]
        if hasattr(obj, "__jailed_pairs__"):
            _id = get_key_by_value(self._store, obj.__jailed_pairs__)
            self.fetch(_id)
        return obj


class Promise(object):  # pylint: disable=useless-object-inheritance
    """Represent a promise."""

    def __init__(self, pfunc):
        """Set up promise."""
        self._resolve_handler = None
        self._finally_handler = None
        self._catch_handler = None

        def resolve(*args, **kwargs):
            self.resolve(*args, **kwargs)

        def reject(*args, **kwargs):
            self.reject(*args, **kwargs)

        pfunc(resolve, reject)

    def resolve(self, result):
        """Resolve promise."""
        try:
            if self._resolve_handler:
                self._resolve_handler(result)
        except Exception as exc:  # pylint: disable=broad-except
            if self._catch_handler:
                self._catch_handler(exc)
            elif not self._finally_handler:
                print("Uncaught Exception: {}".format(exc))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def reject(self, error):
        """Reject promise."""
        try:
            if self._catch_handler:
                self._catch_handler(error)
            elif not self._finally_handler:
                print("Uncaught Exception: {}".format(error))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def then(self, handler):
        """Implement then callback.

        Set handler and return the promise.
        """
        self._resolve_handler = handler
        return self

    def finally_(self, handler):
        """Implement finally callback.

        Set handler and return the promise.
        """
        self._finally_handler = handler
        return self

    def catch(self, handler):
        """Implement catch callback.

        Set handler and return the promise.
        """
        self._catch_handler = handler
        return self



"""Provide a terminal service."""
import asyncio
import os
import select
import shlex
import struct
import subprocess
import sys

from imjoy import __version__
from imjoy.connection.decorator import socketio_handler as sio_on
from imjoy.utils import get_psutil

if sys.platform != "win32":
    import fcntl
    import pty
    import termios


def setup_terminal(engine):
    """Set up the terminal service."""
    engine.conn.register_event_handler(on_start_terminal)
    engine.conn.register_event_handler(on_terminal_input)
    engine.conn.register_event_handler(on_terminal_window_resize)


async def read_and_forward_terminal_output(engine):
    """Read from terminal and forward to the client."""
    terminal_session = engine.store.terminal_session
    max_read_bytes = 1024 * 20
    try:
        terminal_session["output_monitor_running"] = True
        while True:
            await asyncio.sleep(0.01)
            if "fd" in terminal_session:
                timeout_sec = 0
                (data_ready, _, _) = select.select(
                    [terminal_session["fd"]], [], [], timeout_sec
                )
                if data_ready:
                    output = os.read(terminal_session["fd"], max_read_bytes).decode()
                    if output:
                        await engine.conn.sio.emit(
                            "terminal_output", {"output": output}
                        )
    finally:
        terminal_session["output_monitor_running"] = False


@sio_on("start_terminal")
async def on_start_terminal(engine, sid, kwargs):
    """Handle new terminal client connected."""
    if sys.platform == "win32":
        return {"success": False, "error": "Terminal is not available on Windows yet."}
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    terminal_session = engine.store.terminal_session
    try:
        if sid not in registered_sessions:
            logger.debug("Client %s is not registered", sid)
            return {"success": False, "error": "client not registered."}

        if "child_pid" in terminal_session and "fd" in terminal_session:
            process_exists = True
            psutil = get_psutil()
            if psutil is not None:
                process_exists = False
                current_process = psutil.Process()
                children = current_process.children(recursive=True)
                for proc in children:
                    if proc.is_running() and proc.status() != psutil.STATUS_ZOMBIE:
                        if proc.pid == terminal_session["child_pid"]:
                            process_exists = True
                            break
            if process_exists:
                # already started child process, don't start another
                return {
                    "success": True,
                    "exists": True,
                    "message": (
                        f"Welcome to ImJoy Plugin Engine Terminal (v{__version__})."
                    ),
                }

        if sys.platform == "linux" or sys.platform == "linux2":
            # linux
            default_terminal_command = ["bash"]
        elif sys.platform == "darwin":
            # OS X
            default_terminal_command = ["bash"]
        elif sys.platform == "win32":
            # Windows
            default_terminal_command = ["cmd.exe"]
        else:
            default_terminal_command = ["bash"]
        cmd = kwargs.get("cmd", default_terminal_command)

        # create child process attached to a pty we can read from and write to
        (child_pid, fdesc) = pty.fork()
        if child_pid == 0:
            # this is the child process fork.
            # anything printed here will show up in the pty, including the output
            # of this subprocess
            term_env = os.environ.copy()
            term_env["TERM"] = "xterm-256color"
            subprocess.run(cmd, env=term_env)
            subprocess.run(cmd)
        else:
            # this is the parent process fork.
            # store child fd and pid
            terminal_session["fd"] = fdesc
            terminal_session["child_pid"] = child_pid
            set_winsize(fdesc, 50, 50)
            cmd = " ".join(shlex.quote(c) for c in cmd)
            logger.info(
                "Terminal subprocess started, command: %s, pid: %s", cmd, child_pid
            )
            logger.debug("Terminal session %s started", terminal_session)
            if (
                "output_monitor_running" not in terminal_session
                or not terminal_session["output_monitor_running"]
            ):
                asyncio.ensure_future(
                    read_and_forward_terminal_output(engine),
                    loop=asyncio.get_event_loop(),
                )

        return {
            "success": True,
            "message": f"Welcome to ImJoy Plugin Engine Terminal (v{__version__}).",
        }
    except Exception as exc:  # pylint: disable=broad-except
        return {"success": False, "error": str(exc)}


@sio_on("terminal_input")
async def on_terminal_input(engine, sid, data):
    """Write to the terminal as if you are typing in a real terminal."""
    if sys.platform == "win32":
        return "Terminal is not available on Windows yet."

    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    terminal_session = engine.store.terminal_session
    if sid not in registered_sessions:
        return
    try:
        if "fd" in terminal_session:
            os.write(terminal_session["fd"], data["input"].encode())
        else:
            return "Terminal session is closed"
    except Exception as exc:  # pylint: disable=broad-except
        logger.debug("Failed to write to terminal process: %s", exc)
        return str(exc)


def set_winsize(fdesc, row, col, xpix=0, ypix=0):
    """Set window size."""
    if sys.platform == "win32":
        return
    winsize = struct.pack("HHHH", row, col, xpix, ypix)
    fcntl.ioctl(fdesc, termios.TIOCSWINSZ, winsize)


@sio_on("terminal_window_resize")
async def on_terminal_window_resize(engine, sid, data):
    """Resize terminal window."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    terminal_session = engine.store.terminal_session
    if sid not in registered_sessions:
        return
    try:
        if "fd" in terminal_session:
            set_winsize(terminal_session["fd"], data["rows"], data["cols"])
    except Exception as exc:  # pylint: disable=broad-except
        logger.debug("Failed to resize the terminal window: %s", exc)
        return str(exc)



"""Provide worker functions for Python 3."""
import asyncio
import inspect
import logging
import sys
import traceback

import janus

from .utils import format_traceback, Registry
from .utils3 import make_coro
from .python_client import BaseClient, JOB_HANDLERS

# pylint: disable=unused-argument, redefined-outer-name

logger = logging.getLogger("python3_client")

JOB_HANDLERS_PY3 = Registry()
JOB_HANDLERS_PY3.update({name: make_coro(func) for name, func in JOB_HANDLERS.items()})


async def task_worker(conn, async_q, logger, abort=None):
    """Implement a task worker."""
    while True:
        if abort is not None and abort.is_set():
            break
        job = await async_q.get()
        if job is None:
            continue
        handler = JOB_HANDLERS_PY3.get(job["type"])
        if handler is None:
            continue
        try:
            await handler(conn, job, logger)
        except Exception:  # pylint: disable=broad-except
            logger.error("Error occured in the loop %s", traceback.format_exc())
        finally:
            sys.stdout.flush()
            async_q.task_done()


@JOB_HANDLERS_PY3.register("method")
async def handle_method_py3(conn, job, logger):
    """Handle method."""
    if job["name"] in conn.interface:
        if "promise" in job:
            try:
                resolve, reject = conn.unwrap(job["promise"], False)
                method = conn.interface[job["name"]]
                args = conn.unwrap(job["args"], True)
                # args.append({'id': conn.id})
                result = method(*args)
                if result is not None and inspect.isawaitable(result):
                    result = await result
                resolve(result)
            except Exception:  # pylint: disable=broad-except
                traceback_error = traceback.format_exc()
                logger.error("Error in method %s: %s", job["name"], traceback_error)
                reject(Exception(format_traceback(traceback_error)))
        else:
            try:
                method = conn.interface[job["name"]]
                args = conn.unwrap(job["args"], True)
                # args.append({'id': conn.id})
                result = method(*args)
                if result is not None and inspect.isawaitable(result):
                    await result
            except Exception:  # pylint: disable=broad-except
                logger.error(
                    "Error in method %s: %s", job["name"], traceback.format_exc()
                )
    else:
        raise Exception("method " + job["name"] + " is not found.")


@JOB_HANDLERS_PY3.register("callback")
async def handle_callback_py3(conn, job, logger):
    """Handle callback."""
    if "promise" in job:
        resolve, reject = conn.unwrap(job["promise"], False)
        try:
            method = conn.store.fetch(job["num"])
            if method is None:
                raise Exception(
                    "Callback function can only called once, "
                    "if you want to call a function for multiple times, "
                    "please make it as a plugin api function. "
                    "See https://imjoy.io/docs for more details."
                )
            args = conn.unwrap(job["args"], True)
            result = method(*args)
            if result is not None and inspect.isawaitable(result):
                result = await result
            resolve(result)
        except Exception:  # pylint: disable=broad-except
            traceback_error = traceback.format_exc()
            logger.error("Error in method %s: %s", job["num"], traceback_error)
            reject(Exception(format_traceback(traceback_error)))
    else:
        try:
            method = conn.store.fetch(job["num"])
            if method is None:
                raise Exception(
                    "Callback function can only called once, "
                    "if you want to call a function for multiple times, "
                    "please make it as a plugin api function. "
                    "See https://imjoy.io/docs for more details."
                )
            args = conn.unwrap(job["args"], True)
            result = method(*args)
            if result is not None and inspect.isawaitable(result):
                await result
        except Exception:  # pylint: disable=broad-except
            logger.error("Error in method %s: %s", job["num"], traceback.format_exc())


class AsyncClient(BaseClient):
    """Represent an async socketio client."""

    # pylint: disable=too-few-public-methods

    def __init__(self, id=None):
        """Set up client instance."""
        super().__init__(id)
        self.loop = asyncio.get_event_loop()
        self.janus_queue = janus.Queue(loop=self.loop)
        self.queue = self.janus_queue.sync_q
        self.task_worker = task_worker

    def run_forever(self, conn):
        """Run forever."""

        if self.loop.is_running():
            for _ in range(10):
                self.loop.create_task(
                    self.task_worker(conn, self.janus_queue.async_q, logger, conn.abort)
                )
        else:
            workers = [
                self.task_worker(conn, self.janus_queue.async_q, logger, conn.abort)
                for i in range(10)
            ]
            self.loop.run_until_complete(asyncio.gather(*workers))



"""Provide worker functions."""
import logging
import traceback
import sys
import time
import uuid

import socketio

from .utils import format_traceback, Registry

try:
    import queue
except ImportError:
    import Queue as queue

# pylint: disable=unused-argument, redefined-outer-name

JOB_HANDLERS = Registry()
logger = logging.getLogger("python_client")


def task_worker(conn, sync_q, logger, abort):
    """Implement a task worker."""
    while True:
        if abort is not None and abort.is_set():
            break
        try:
            job = sync_q.get()
        except queue.Empty:
            time.sleep(0.1)
            continue
        sync_q.task_done()
        if job is None:
            continue
        handler = JOB_HANDLERS.get(job["type"])
        if handler is None:
            continue

        try:
            handler(conn, job, logger)
        except Exception:  # pylint: disable=broad-except
            logger.error("Error occured in the loop %s", traceback.format_exc())
        sys.stdout.flush()


@JOB_HANDLERS.register("getInterface")
def handle_get_interface(conn, job, logger):
    """Handle get interface."""
    conn.send_interface()


@JOB_HANDLERS.register("setInterface")
def handle_set_interface(conn, job, logger):
    """Handle set interface."""
    conn.set_remote(job["api"])
    conn.emit({"type": "interfaceSetAsRemote"})
    if not conn.init:
        conn.emit({"type": "getInterface"})
        conn.init = True


@JOB_HANDLERS.register("interfaceSetAsRemote")
def handle_set_interface_as_remote(conn, job, logger):
    """Handle set interface as remote."""
    # conn.emit({'type':'getInterface'})
    conn.remote_set = True


@JOB_HANDLERS.register("execute")
def handle_execute(conn, job, logger):
    """Handle execute."""
    if not conn.executed:
        try:
            type_ = job["code"]["type"]
            if type_ == "script":
                content = job["code"]["content"]
                exec(content, conn.local)  # pylint: disable=exec-used
                conn.executed = True
            elif type_ == "requirements":
                pass
            else:
                raise Exception("unsupported type")
            conn.emit({"type": "executeSuccess"})
        except Exception:  # pylint: disable=broad-except
            traceback_error = traceback.format_exc()
            logger.error("Error during execution: %s", traceback_error)
            conn.emit({"type": "executeFailure", "error": traceback_error})
    else:
        logger.info("Skip code execution")


@JOB_HANDLERS.register("method")
def handle_method(conn, job, logger):
    """Handle method."""
    interface = conn.interface
    if "pid" in job and job["pid"] is not None:
        interface = conn.plugin_interfaces[job["pid"]]
    if job["name"] in interface:
        if "promise" in job:
            try:
                resolve, reject = conn.unwrap(job["promise"], False)
                method = interface[job["name"]]
                args = conn.unwrap(job["args"], True)
                # args.append({'id': conn.id})
                result = method(*args)
                resolve(result)
            except Exception:  # pylint: disable=broad-except
                traceback_error = traceback.format_exc()
                logger.error("Error in method %s: %s", job["name"], traceback_error)
                reject(Exception(format_traceback(traceback_error)))
        else:
            try:
                method = interface[job["name"]]
                args = conn.unwrap(job["args"], True)
                # args.append({'id': conn.id})
                method(*args)
            except Exception:  # pylint: disable=broad-except
                logger.error(
                    "Error in method %s: %s", job["name"], traceback.format_exc()
                )
    else:
        raise Exception("method " + job["name"] + " is not found.")


@JOB_HANDLERS.register("callback")
def handle_callback(conn, job, logger):
    """Handle callback."""
    if "promise" in job:
        resolve, reject = conn.unwrap(job["promise"], False)
        try:
            method = conn.store.fetch(job["num"])
            if method is None:
                raise Exception(
                    "Callback function can only called once, "
                    "if you want to call a function for multiple times, "
                    "please make it as a plugin api function. "
                    "See https://imjoy.io/docs for more details."
                )
            args = conn.unwrap(job["args"], True)
            result = method(*args)
            resolve(result)
        except Exception:  # pylint: disable=broad-except
            traceback_error = traceback.format_exc()
            logger.error("Error in method %s: %s", job["num"], traceback_error)
            reject(Exception(format_traceback(traceback_error)))
    else:
        try:
            method = conn.store.fetch(job["num"])
            if method is None:
                raise Exception(
                    "Callback function can only called once, "
                    "if you want to call a function for multiple times, "
                    "please make it as a plugin api function. "
                    "See https://imjoy.io/docs for more details."
                )
            args = conn.unwrap(job["args"], True)
            method(*args)
        except Exception:  # pylint: disable=broad-except
            logger.error("Error in method %s: %s", job["num"], traceback.format_exc())


class BaseClient(object):  # pylint: disable=useless-object-inheritance
    """Represent a base socketio client."""

    queue = None
    _clients = {}

    @staticmethod
    def get_client(id):
        return BaseClient._clients.get(id)

    def __init__(self, id=None):
        """Set up client instance."""
        self.id = id or str(uuid.uuid4())
        self.sio = socketio.Client()
        BaseClient._clients[self.id] = self

    def setup(self, conn):
        """Set up the plugin connection."""
        self.sio.connect(conn.opt.server)

        def emit(msg):
            """Emit a message to the socketio server."""
            self.sio.emit("from_plugin_" + conn.secret, msg)

        def sio_plugin_message(*args):
            """Handle plugin message."""
            data = args[0]
            if data["type"] == "import":
                emit({"type": "importSuccess", "url": data["url"]})
            elif data["type"] == "disconnect":
                conn.abort.set()
                try:
                    if "exit" in conn.interface and callable(conn.interface["exit"]):
                        conn.interface["exit"]()
                except Exception as exc:  # pylint: disable=broad-except
                    logger.error("Error when exiting: %s", exc)

            elif data["type"] == "execute":
                if not conn.executed:
                    self.queue.put(data)
                else:
                    logger.debug("Skip execution")
                    emit({"type": "executeSuccess"})
            elif data["type"] == "message":
                _data = data["data"]
                self.queue.put(_data)
                logger.debug("Added task to the queue")

        def on_disconnect():
            if not conn.opt.daemon:
                conn.exit(1)

        conn.emit = emit
        self.sio.on("disconnect", on_disconnect)
        self.sio.on("to_plugin_" + conn.secret, sio_plugin_message)
        emit({"type": "initialized", "dedicatedThread": True})
        logger.info("Plugin %s initialized", conn.opt.id)

    def run_forever(self, conn):
        """Run forever."""
        raise NotImplementedError


class Client(BaseClient):
    """Represent a sync socketio client."""

    def __init__(self):
        """Set up client instance."""
        super(Client, self).__init__()
        self.queue = queue.Queue()
        self.task_worker = task_worker

    def run_forever(self, conn):
        """Run forever."""
        self.task_worker(conn, self.queue, logger, conn.abort)



"""Provide project scoped test fixtures."""
import signal
import subprocess
import time
import uuid

import pytest

from tests.common import ENGINE_MODULE, TOKEN, URL
from tests.mock_client import TestClient


@pytest.fixture(name="engine")
def setup_engine():
    """Set up engine."""
    engine_args = f"python {ENGINE_MODULE} --dev --debug --token {TOKEN}"
    process = subprocess.Popen(engine_args.split())
    time.sleep(2)  # This is needed to let the engine finish setup.
    yield
    process.send_signal(signal.SIGINT)
    try:
        process.wait(timeout=10)
    except subprocess.TimeoutExpired:
        print("Waiting for engine exit timed out")


@pytest.fixture(name="client")
async def mock_client(engine, event_loop):  # pylint: disable=unused-argument
    """Provide a mock client."""
    client_id = str(uuid.uuid4())
    session_id = str(uuid.uuid4())
    client = TestClient(URL, client_id, session_id, TOKEN, event_loop)
    await client.connect()
    await client.register_client()
    yield client
    await client.sio.disconnect()



"""Provide the socketio connection manager."""
import asyncio
from functools import partial

import socketio

from imjoy.utils import dotdict
from .decorator import partial_coro
from .handler import register_services
from .server import create_app, run_app


def create_connection_manager(engine):
    """Create a socketio connection and return the connection instance."""
    # ALLOWED_ORIGINS = [opt.base_url, 'http://imjoy.io', 'https://imjoy.io']
    sio = socketio.AsyncServer(cors_allowed_origins="*")
    app = create_app(engine)
    sio.attach(app)
    return ConnectionManager(engine, app, sio)


def register_event_handler(engine, event, handler=None, namespace=None):
    """Register a socketio event handler."""
    # An event handler can be found like this:
    # handler = sio.handlers[namespace][event]
    # pylint: disable=protected-access
    if handler is None:
        handler = event
        event = handler._ws_event
        namespace = handler._ws_namespace
    if asyncio.iscoroutinefunction(handler):
        injected_handler = partial_coro(handler, engine)
    else:
        injected_handler = partial(handler, engine)
    engine.conn.sio.on(event, handler=injected_handler, namespace=namespace)


class ConnectionManager:
    """Represent a connection manager for socketio event handler and session data."""

    def __init__(self, engine, app, sio):
        """Set up connection instance attributes."""
        self.engine = engine
        self.app = app
        self.sio = sio
        self.store = dotdict()
        self.reset_store()

    def reset_store(self, reset_clients=True):
        """Reset the connection data store."""
        self.store.attempt_count = 0
        self.store.cmd_history = []
        self.store.plugins = {}
        self.store.plugin_sessions = {}
        self.store.plugin_sids = {}
        self.store.plugin_signatures = {}

        self.store.generated_urls = {}
        self.store.generated_url_files = {}
        self.store.request_upload_files = {}
        self.store.request_urls = {}
        self.store.terminal_session = {}

        if reset_clients:
            self.store.clients = {}
            self.store.client_sessions = {}
            self.store.registered_sessions = {}

    def setup(self):
        """Set up the connection manager."""
        self._register_services()

    def start(self):
        """Start the connection."""
        run_app(self.engine, self.app)

    def register_event_handler(self, event, handler=None, namespace=None):
        """Register a socketio event handler."""
        register_event_handler(self.engine, event, handler=handler, namespace=namespace)

    def _register_services(self):
        """Register event handlers for internal services."""
        register_services(self.engine, register_event_handler)



"""Test plugin engine api."""
import pytest
import asyncio

import numpy as np

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio  # pylint: disable=invalid-name

TEST_PLUGIN_CONFIG = {
    "name": "test-plugin",
    "type": "native-python",
    "version": "0.1.12",
    "api_version": "0.1.2",
    "description": "This is a test plugin.",
    "tags": ["CPU", "GPU", "macOS CPU"],
    "ui": "",
    "inputs": None,
    "outputs": None,
    "flags": [],
    "icon": None,
    "env": "",
    "requirements": "",
    "dependencies": [],
}

TEST_PLUGIN_SCRIPT = """
import asyncio
import numpy as np
from imjoy import api

class ImJoyPlugin():
    def setup(self):
        api.log('initialized')

    async def run(self, ctx):
        await api.alert('Hello')
        await api.log('done')

    def get_array(self):
        array = np.zeros(10)
        # array[3] = 93834.356
        return {'array': array}

api.export(ImJoyPlugin())
"""


@pytest.fixture(name="test_plugin")
async def setup_test_plugin(client, event_loop):
    """Initialize the plugin."""
    plugin = await client.init_plugin(TEST_PLUGIN_CONFIG)
    return plugin


@pytest.fixture(name="test_plugin_executed")
async def test_plugin_execute(client, test_plugin, event_loop):
    """Test plugin execute."""
    await test_plugin.execute({"type": "script", "content": TEST_PLUGIN_SCRIPT})
    return test_plugin


async def test_plugin_run(client, test_plugin_executed, event_loop):
    """Test run function."""
    api = test_plugin_executed.get_api()
    await api.run({})


async def test_numpy_plugin(client, test_plugin_executed, event_loop):
    """Test get numpy array."""
    api = test_plugin_executed.get_api()
    result = await api.get_array()
    assert np.array_equal(result["array"], np.zeros(10))



"""Package the ImJoy plugin engine."""
import json
import os

# read version information from file
IMJOY_PACKAGE_DIR = os.path.dirname(__file__)
with open(os.path.join(IMJOY_PACKAGE_DIR, "VERSION"), "r") as f:
    VERSION_INFO = json.load(f)
    __version__ = VERSION_INFO["version"]
    API_VERSION = VERSION_INFO["api_version"]



"""Test the testing framework."""
import pytest


def test_framework_pass():
    """Test that the testing framework works and a test can pass."""
    assert True


@pytest.mark.xfail
def test_framework_fail():
    """Test that the testing framework works and a test can fail."""
    assert False



"""Provide common utils for tests."""
from pathlib import Path

HERE = Path(__file__).parent
ENGINE_MODULE = HERE.parent / "imjoy/engine.py"
TOKEN = "12345678"
HOST = "localhost"
PORT = 9527
URL = f"http://{HOST}:{PORT}"



"""Provide socketio event handlers."""
from imjoy.utils import get_psutil

from .decorator import socketio_handler as sio_on

NAME_SPACE = "/"


def register_services(engine, register_event_handler):
    """Register services running by the engine."""
    # basic engine service
    register_event_handler(engine, connect)
    register_event_handler(engine, disconnect)
    register_event_handler(engine, reset_engine)
    register_event_handler(engine, on_get_engine_status)


@sio_on("connect")
def connect(engine, sid, _):
    """Connect client."""
    logger = engine.logger
    logger.info("Connect %s", sid)


@sio_on("reset_engine")
async def reset_engine(engine, sid, kwargs):
    """Reset engine."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered"}

    # services and runners can register reset_engine_* handlers that will be called here
    for event, handler in engine.conn.sio.handlers[NAME_SPACE].items():
        if not event.startswith("reset_engine_"):
            continue
        await handler(sid, kwargs)

    engine.conn.reset_store(reset_clients=False)

    return {"success": True}


@sio_on("get_engine_status")
async def on_get_engine_status(engine, sid, _):
    """Return engine status."""
    logger = engine.logger
    plugins = engine.store.plugins
    registered_sessions = engine.store.registered_sessions
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered."}
    psutil = get_psutil()
    if psutil is None:
        return {"success": False, "error": "psutil is not available."}
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    pid_dict = {}
    for plugin in plugins.values():
        if plugin["process_id"] is not None:
            pid_dict[plugin["process_id"]] = plugin

    procs = []
    for proc in children:
        if proc.is_running() and proc.status() != psutil.STATUS_ZOMBIE:
            if proc.pid in pid_dict:
                procs.append({"name": pid_dict[proc.pid]["name"], "pid": proc.pid})
            else:
                procs.append({"name": proc.name(), "pid": proc.pid})

    return {
        "success": True,
        "plugin_num": len(plugins),
        "plugin_processes": procs,
        "engine_process": current_process.pid,
    }


@sio_on("disconnect")
async def disconnect(engine, sid):
    """Disconnect client."""
    logger = engine.logger
    # services and runners can register disconnect_* handlers that will be called here
    for event, handler in engine.conn.sio.handlers[NAME_SPACE].items():
        if not event.startswith("disconnect_"):
            continue
        await handler(sid)
    logger.info("Client(%s) disconnected", sid)



"""Provide a file server service."""
import os
import shutil
import sys
import uuid
from urllib.parse import urlparse

from imjoy.connection.decorator import socketio_handler as sio_on
from imjoy.utils import scandir

if sys.platform == "win32":
    from imjoy.utils import get_drives


def setup_file_server(engine):
    """Set up the file server service."""
    engine.conn.register_event_handler(on_list_dir)
    engine.conn.register_event_handler(on_get_file_url)
    engine.conn.register_event_handler(on_get_file_path)
    engine.conn.register_event_handler(on_remove_files)
    engine.conn.register_event_handler(on_request_upload_url)


@sio_on("list_dir")
async def on_list_dir(engine, sid, kwargs):
    """List files in directory."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered."}

    try:
        workspace_dir = os.path.join(
            engine.opt.workspace_dir, registered_sessions[sid]["workspace"]
        )

        path = kwargs.get("path", workspace_dir)
        path = os.path.normpath(os.path.expanduser(path))
        if not os.path.isabs(path):
            path = os.path.join(workspace_dir, path)
        path = os.path.abspath(path)

        type_ = kwargs.get("type")
        recursive = kwargs.get("recursive", False)
        files_list = {"success": True}
        files_list["sep"] = os.sep
        files_list["path"] = path
        files_list["name"] = os.path.basename(os.path.abspath(path))
        files_list["type"] = "dir"
        files_list["children"] = scandir(files_list["path"], type_, recursive)

        if sys.platform == "win32":
            files_list["drives"] = get_drives()

        return files_list
    except Exception as exc:  # pylint: disable=broad-except
        logger.error("List dir error: %s", exc)
        return {"success": False, "error": str(exc)}


@sio_on("remove_files")
async def on_remove_files(engine, sid, kwargs):
    """Remove files."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered."}
    logger.info("Removing files: %s", kwargs)
    workspace_dir = os.path.join(
        engine.opt.workspace_dir, registered_sessions[sid]["workspace"]
    )
    path = kwargs.get("path", workspace_dir)
    if not os.path.isabs(path):
        path = os.path.join(workspace_dir, path)
    path = os.path.normpath(os.path.expanduser(path))
    type_ = kwargs.get("type")
    recursive = kwargs.get("recursive", False)

    if os.path.exists(path) and not os.path.isdir(path) and type_ == "file":
        try:
            os.remove(path)
            return {"success": True}
        except Exception as exc:  # pylint: disable=broad-except
            logger.error("Remove files error: %s", str(exc))
            return {"success": False, "error": str(exc)}
    elif os.path.exists(path) and os.path.isdir(path) and type_ == "dir":
        try:
            if recursive:
                dirname, filename = os.path.split(path)
                shutil.move(path, os.path.join(dirname, "." + filename))
                # shutil.rmtree(path)
            else:
                os.rmdir(path)
            return {"success": True}
        except Exception as exc:  # pylint: disable=broad-except
            logger.error("Remove files error: %s", str(exc))
            return {"success": False, "error": str(exc)}
    else:
        logger.error("Remove files error: File does not exists or type mismatch")
        return {"success": False, "error": "File not exists or type mismatch."}


@sio_on("request_upload_url")
async def on_request_upload_url(engine, sid, kwargs):
    """Request upload url."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    request_upload_files = engine.store.request_upload_files
    request_urls = engine.store.request_urls
    logger.info("Requesting file upload url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered"}

    urlid = str(uuid.uuid4())
    file_info = {
        "id": urlid,
        "overwrite": kwargs.get("overwrite", False),
        "workspace": registered_sessions[sid]["workspace"],
    }
    if "path" in kwargs:
        file_info["path"] = kwargs["path"]

    if "dir" in kwargs:
        path = os.path.expanduser(kwargs["dir"])
        if not os.path.isabs(path):
            path = os.path.join(engine.opt.workspace_dir, file_info["workspace"], path)
        file_info["dir"] = path

    if "path" in file_info:
        path = file_info["path"]
        if "dir" in file_info:
            path = os.path.join(file_info["dir"], path)
        else:
            path = os.path.join(engine.opt.workspace_dir, file_info["workspace"], path)

        if os.path.exists(path) and not kwargs.get("overwrite", False):
            return {"success": False, "error": "file already exist."}

    base_url = kwargs.get("base_url", registered_sessions[sid]["base_url"])
    url = "{}/upload/{}".format(base_url, urlid)
    request_urls[url] = file_info
    request_upload_files[urlid] = file_info
    return {"success": True, "id": urlid, "url": url}


@sio_on("get_file_url")
async def on_get_file_url(engine, sid, kwargs):
    """Return file url."""
    logger = engine.logger
    generated_url_files = engine.store.generated_url_files
    generated_urls = engine.store.generated_urls
    registered_sessions = engine.store.registered_sessions
    logger.info("Generating file url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered"}

    path = os.path.abspath(os.path.expanduser(kwargs["path"]))
    if not os.path.exists(path):
        return {"success": False, "error": "file does not exist."}
    file_info = {"path": path}
    if os.path.isdir(path):
        file_info["type"] = "dir"
    else:
        file_info["type"] = "file"
    if kwargs.get("headers"):
        file_info["headers"] = kwargs["headers"]
    _, name = os.path.split(path)
    file_info["name"] = name
    if path in generated_url_files:
        return {"success": True, "url": generated_url_files[path]}
    urlid = str(uuid.uuid4())
    generated_urls[urlid] = file_info
    base_url = kwargs.get("base_url", registered_sessions[sid]["base_url"])
    if kwargs.get("password"):
        file_info["password"] = kwargs["password"]
        generated_url_files[path] = "{}/file/{}@{}/{}".format(
            base_url, urlid, file_info["password"], name
        )
    else:
        generated_url_files[path] = "{}/file/{}/{}".format(base_url, urlid, name)
    return {"success": True, "url": generated_url_files[path]}


@sio_on("get_file_path")
async def on_get_file_path(engine, sid, kwargs):
    """Return file path."""
    logger = engine.logger
    generated_urls = engine.store.generated_urls
    registered_sessions = engine.store.registered_sessions
    logger.info("Generating file url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered"}

    url = kwargs["url"]
    urlid = urlparse(url).path.replace("/file/", "")
    if urlid in generated_urls:
        file_info = generated_urls[urlid]
        return {"success": True, "path": file_info["path"]}
    return {"success": False, "error": "url not found."}



"""Implement the ImJoy plugin engine."""
import logging
import sys

from imjoy.options import parse_cmd_line
from imjoy.env import bootstrap, prepare_env
from imjoy.connection import create_connection_manager
from imjoy.runners import setup_runners
from imjoy.services import setup_services


def setup_logging(opt, logger):
    """Set up logging."""
    if opt.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


class Engine:
    """Represent the plugin engine."""

    def __init__(self, opt, logger):
        """Set up instance attributes of the engine."""
        self.logger = logger
        self.opt = opt
        self.conn = None
        self.store = None

    def __repr__(self):
        """Return the engine representation."""
        return f"<Engine(opt={self.opt})>"

    @classmethod
    def create(cls, args):
        """Prepare and and return an Engine instance."""
        logging.basicConfig(stream=sys.stdout)
        logger = logging.getLogger("ImJoyPluginEngine")
        opt = parse_cmd_line(args)
        setup_logging(opt, logger)
        opt = prepare_env(opt, logger)
        opt = bootstrap(opt, logger)
        return cls(opt, logger)

    def setup(self):
        """Set up the engine."""
        self.conn = create_connection_manager(self)
        self.store = self.conn.store
        self.conn.setup()
        setup_services(self)
        setup_runners(self)

    def start(self):
        """Start the engine."""
        self.conn.start()


def run(args=None):
    """Run the engine."""
    engine = Engine.create(args)
    engine.setup()
    engine.start()


if __name__ == "__main__":
    run()



"""Provide engine services."""
from .file_server import setup_file_server
from .terminal import setup_terminal


def setup_services(engine):
    """Set up engine services."""
    setup_file_server(engine)
    setup_terminal(engine)



"""Provide setup function to prepare the engine."""
import json
import os
import shutil
import subprocess
import sys
import time
import uuid

from imjoy.utils import kill_process, read_or_generate_token


def prepare_env(opt, logger):
    """Prepare environment."""
    opt.conda_available = False
    # add executable path to PATH
    os.environ["PATH"] = (
        os.path.split(sys.executable)[0] + os.pathsep + os.environ.get("PATH", "")
    )

    try:
        process = subprocess.Popen(
            ["conda", "info", "--json", "-s"], stdout=subprocess.PIPE
        )
        cout, _ = process.communicate()
        conda_prefix = json.loads(cout.decode("ascii"))["conda_prefix"]
        logger.info("Found conda environment: %s", conda_prefix)
        # for fixing CondaHTTPError:
        # https://github.com/conda/conda/issues/6064#issuecomment-458389796
        if os.name == "nt":
            os.environ["PATH"] = (
                os.path.join(conda_prefix, "Library", "bin")
                + os.pathsep
                + os.environ["PATH"]
            )
        opt.conda_available = True
    except OSError:
        conda_prefix = None
        if sys.version_info < (3, 0):
            sys.exit(
                "Sorry, ImJoy plugin engine can only run within a conda environment "
                "or at least in Python 3."
            )
        logger.warning(
            "You are running ImJoy without conda, "
            "you may have problems with some plugins"
        )

    if opt.conda_available:
        if sys.platform == "linux" or sys.platform == "linux2":
            # linux
            opt.conda_activate = (
                "/bin/bash -c 'source " + conda_prefix + "/bin/activate {}'"
            )
        elif sys.platform == "darwin":
            # OS X
            opt.conda_activate = "source activate {}"
        elif sys.platform == "win32":
            # Windows...
            opt.conda_activate = "activate {}"
        else:
            opt.conda_activate = "conda activate {}"
    else:
        opt.conda_activate = "{}"

    return opt


def bootstrap(opt, logger):
    """Bootstrap the engine."""
    if not opt.conda_available and not opt.freeze:
        logger.warning(
            "Command `pip install` may not work, "
            "in that case you may want to add `--freeze`"
        )

    if opt.freeze:
        logger.warning(
            "You are running the plugin engine with `--freeze`, "
            "this means you need to handle all the plugin requirements yourself"
        )

    opt.workspace_dir = os.path.expanduser(opt.workspace)
    if not os.path.exists(opt.workspace_dir):
        os.makedirs(opt.workspace_dir)
    if opt.token is None or opt.token == "":
        opt.token = read_or_generate_token(
            os.path.join(opt.workspace_dir, ".token"), logger
        )

    # try to kill last process
    pid_file = os.path.join(opt.workspace_dir, ".pid")
    try:
        if os.path.exists(pid_file):
            with open(pid_file, "r") as fil:
                pid = int(fil.read())
                logger.info("Trying to kill last process (pid=%s)", pid)
                kill_process(pid, logger)
                # wait for a while to release the port
                time.sleep(3)

    except Exception:  # pylint: disable=broad-except
        logger.debug("Failed to kill last process")
    try:
        engine_pid = str(os.getpid())
        with open(pid_file, "w") as fil:
            fil.write(engine_pid)
    except Exception as exc:  # pylint: disable=broad-except
        logger.error("Failed to save .pid file: %s", str(exc))

    opt.web_app_dir = os.path.join(opt.workspace_dir, "__ImJoy__")
    if opt.serve:
        if shutil.which("git") is None:
            logger.info("Installing git")
            ret = subprocess.Popen(
                "conda install -y git && git clone -b gh-pages --depth 1 "
                "https://github.com/oeway/ImJoy".split(),
                shell=False,
            ).wait()
            if ret != 0:
                logger.error(
                    "Failed to install git, "
                    "please check whether you have internet access"
                )
                sys.exit(3)
        if os.path.exists(opt.web_app_dir) and os.path.isdir(opt.web_app_dir):
            ret = subprocess.Popen(
                ["git", "stash"], cwd=opt.web_app_dir, shell=False
            ).wait()
            if ret != 0:
                logger.error("Failed to clean files locally")
            ret = subprocess.Popen(
                ["git", "pull", "--all"], cwd=opt.web_app_dir, shell=False
            ).wait()
            if ret != 0:
                logger.error("Failed to pull files for serving offline")
            ret = subprocess.Popen(
                ["git", "checkout", "gh-pages"], cwd=opt.web_app_dir, shell=False
            ).wait()
            if ret != 0:
                logger.error("Failed to checkout files from gh-pages")
        if not os.path.exists(opt.web_app_dir):
            logger.info("Downloading files for serving ImJoy locally")
            ret = subprocess.Popen(
                "git clone -b gh-pages --depth 1 "
                "https://github.com/oeway/ImJoy __ImJoy__".split(),
                shell=False,
                cwd=opt.workspace_dir,
            ).wait()
            if ret != 0:
                logger.error(
                    "Failed to download files, "
                    "please check whether you have internet access"
                )
                sys.exit(4)

    return opt



"""Provide a web server."""
import json
import os
import sys
import threading
import time
from mimetypes import MimeTypes

import aiohttp_cors
from aiohttp import web, streamer

from imjoy import __version__, API_VERSION
from imjoy.utils import kill_process, scandir

ENGINE = "imjoy_engine"


def create_app(engine):
    """Create and return aiohttp webserver app."""
    app = web.Application()
    app[ENGINE] = engine
    setup_router(engine, app)
    setup_cors(app)
    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    return app


def run_app(engine, app):
    """Run the app."""
    logger = engine.logger
    try:
        web.run_app(app, host=engine.opt.host, port=int(engine.opt.port))
    except OSError as exc:
        if exc.errno in {48}:
            logger.error(
                "Failed to open port %s for ImJoy Engine, "
                "please try to terminate the process which is using that port, "
                "or restart your computer.",
                engine.opt.port,
            )
        else:
            logger.error("Failed to start ImJoy Engine, error: %s", exc)
    except Exception as e:
        logger.error("Failed to start ImJoy Engine, error: %s", e)


def setup_router(engine, app):
    """Set up router."""
    # pylint: disable=unused-argument
    logger = engine.logger
    if engine.opt.serve and os.path.exists(
        os.path.join(engine.opt.web_app_dir, "index.html")
    ):

        async def index(request):
            """Serve the client-side application."""
            with open(
                os.path.join(engine.opt.web_app_dir, "index.html"),
                "r",
                encoding="utf-8",
            ) as fil:
                return web.Response(text=fil.read(), content_type="text/html")

        app.router.add_static(
            "/static", path=str(os.path.join(engine.opt.web_app_dir, "static"))
        )
        # app.router.add_static('/docs/', path=str(os.path.join(web_app_dir, 'docs')))

        async def docs_handler(request):
            """Handle docs."""
            raise web.HTTPFound(location="https://imjoy.io/docs")

        app.router.add_get("/docs", docs_handler, name="docs")
        logger.info(
            "A local version of Imjoy web app is available at %s", engine.opt.base_url
        )
    else:

        async def index(request):
            """Return index."""
            return web.Response(
                body=(
                    '<H1><a href="https://imjoy.io">ImJoy.IO</a></H1><p>'
                    'You can run "python -m imjoy --serve" '
                    "to serve ImJoy web app locally.</p>"
                ),
                content_type="text/html",
            )

    app.router.add_get("/", index)
    app.router.add_get("/about", about)


def setup_cors(app):
    """Set up cors."""
    cors = aiohttp_cors.setup(
        app,
        defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True, expose_headers="*", allow_headers="*"
            )
        },
    )

    # app.router.add_post('/upload/{urlid}', upload_file)
    cors.add(app.router.add_route("POST", "/upload/{urlid}", upload_file))
    cors.add(app.router.add_get("/file/{urlid}/{name:.+}", download_file))
    cors.add(app.router.add_get("/file/{urlid}@{password}/{name:.+}", download_file))


@streamer
async def file_sender(writer, file_path=None):
    """Read a large file chunk by chunk and send it through HTTP.

    Do not read the chunks into memory.
    """
    with open(file_path, "rb") as f:
        chunk = f.read(2 ** 16)
        while chunk:
            await writer.write(chunk)
            chunk = f.read(2 ** 16)


async def about(request):
    """Return about text."""
    engine = request.app[ENGINE]
    params = request.rel_url.query
    if "token" in params:
        body = (
            "<H1>ImJoy Plugin Engine connection token: </H1><H3>"
            + params["token"]
            + "</H3><br>"
        )
        body += (
            "<p>You have to specify this token when you connect the ImJoy web app "
            "to this Plugin Engine. The token will be saved and automatically reused "
            "when you launch the App again. </p>"
        )
        body += "<br>"
        body += (
            "<p>Alternatively, you can launch a new ImJoy instance "
            "with the link below: </p>"
        )

        if engine.opt.serve:
            body += (
                '<p><a href="'
                + engine.opt.base_url
                + "/#/app?token="
                + params["token"]
                + '">Open ImJoy App</a></p>'
            )
        else:
            body += (
                '<p><a href="https://imjoy.io/#/app?token='
                + params["token"]
                + '">Open ImJoy App</a></p>'
            )

    else:
        if engine.opt.serve:
            body = (
                '<H1><a href="'
                + engine.opt.base_url
                + '/#/app">Open ImJoy App</a></H1>'
            )
        else:
            body = '<H1><a href="https://imjoy.io/#/app">Open ImJoy App</a></H1>'
    body += (
        "<H2>Please use the latest Google Chrome browser to run the ImJoy App."
        '</H2><a href="https://www.google.com/chrome/">Download Chrome</a><p>'
        "Note: Safari is not supported "
        "due to its restrictions on connecting to localhost. "
        "Currently, only FireFox and Chrome (preferred) are supported.</p>"
    )
    return web.Response(body=body, content_type="text/html")


async def upload_file(request):
    """Upload file."""
    engine = request.app[ENGINE]
    logger = engine.logger
    request_upload_files = engine.store.request_upload_files
    urlid = request.match_info["urlid"]  # Could be a HUGE file
    if urlid not in request_upload_files:
        raise web.HTTPForbidden(text="Invalid URL")

    file_info = request_upload_files[urlid]
    try:
        reader = await request.multipart()
        field = None
        while True:
            part = await reader.next()
            logger.debug("Reading part %s of %s", part, part.filename)
            if part.filename is None:
                continue
            field = part
            break
        filename = field.filename
        # You cannot rely on Content-Length if transfer is chunked.
        size = 0
        if "path" in file_info:
            path = file_info["path"]
        else:
            path = filename

        if "dir" in file_info:
            path = os.path.join(file_info["dir"], path)
        else:
            path = os.path.join(engine.opt.workspace_dir, file_info["workspace"], path)

        if os.path.exists(path) and not file_info.get("overwrite", False):
            return web.Response(body="File {} already exists.".format(path), status=404)

        logger.info("Uploading file to %s", path)
        directory, _ = os.path.split(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(path, "wb") as fil:
            while True:
                chunk = await field.read_chunk()  # 8192 bytes by default.
                if not chunk:
                    break
                size += len(chunk)
                fil.write(chunk)
        file_info["size"] = size
        file_info["path"] = path
        logger.info("File saved to %s (size %d)", path, size)
        return web.json_response(file_info)

    except Exception as exc:  # pylint: disable=broad-except
        logger.error("Failed to upload file, error: %s", exc)
        return web.Response(body=f"Failed to upload, error: {exc}", status=404)


async def download_file(request):  # pylint: disable=too-many-return-statements
    """Download file."""
    engine = request.app[ENGINE]
    generated_urls = engine.store.generated_urls
    urlid = request.match_info["urlid"]  # Could be a HUGE file
    name = request.match_info["name"]
    if urlid not in generated_urls:
        raise web.HTTPForbidden(text="Invalid URL")
    file_info = generated_urls[urlid]
    if file_info.get("password", False):
        password = request.match_info.get("password")
        if password != file_info["password"]:
            raise web.HTTPForbidden(text="Incorrect password for accessing this file.")
    headers = file_info.get("headers")
    default_headers = {}
    if file_info["type"] == "dir":
        dirname = os.path.dirname(name)
        # list the folder
        if dirname == "" or dirname is None:
            if name != file_info["name"]:
                raise web.HTTPForbidden(text="File name does not match server record!")
            folder_path = file_info["path"]
            if not os.path.exists(folder_path):
                return web.Response(
                    body="Folder <{folder_path}> does not exist".format(
                        folder_path=folder_path
                    ),
                    status=404,
                )

            file_list = scandir(folder_path, "file", False)
            headers = headers or {
                "Content-Disposition": 'inline; filename="{filename}"'.format(
                    filename=name
                )
            }
            headers.update(default_headers)
            return web.json_response(file_list, headers=headers)
        # list the subfolder or get a file in the folder

        file_path = os.path.join(file_info["path"], os.sep.join(name.split("/")[1:]))
        if not os.path.exists(file_path):
            return web.Response(
                body="File <{file_path}> does not exist".format(file_path=file_path),
                status=404,
            )
        if os.path.isdir(file_path):
            _, folder_name = os.path.split(file_path)
            file_list = scandir(file_path, "file", False)
            headers = headers or {
                "Content-Disposition": 'inline; filename="{filename}"'.format(
                    filename=folder_name
                )
            }
            headers.update(default_headers)
            return web.json_response(file_list, headers=headers)

        _, file_name = os.path.split(file_path)
        mime_type = MimeTypes().guess_type(file_name)[0] or "application/octet-stream"
        file_size = os.path.getsize(file_path)
        headers = headers or {
            "Content-Disposition": 'inline; filename="{filename}"'.format(
                filename=file_name
            ),
            "Content-Type": mime_type,
            "Content-Length": str(file_size),
        }
        headers.update(default_headers)
        return web.Response(body=file_sender(file_path=file_path), headers=headers)
    if file_info["type"] == "file":
        file_path = file_info["path"]
        if name != file_info["name"]:
            raise web.HTTPForbidden(text="File name does not match server record!")
        file_name = file_info["name"]
        if not os.path.exists(file_path):
            return web.Response(
                body="File <{file_name}> does not exist".format(file_name=file_path),
                status=404,
            )
        mime_type = MimeTypes().guess_type(file_name)[0] or "application/octet-stream"
        file_size = os.path.getsize(file_path)
        headers = headers or {
            "Content-Disposition": 'inline; filename="{filename}"'.format(
                filename=file_name
            ),
            "Content-Type": mime_type,
            "Content-Length": str(file_size),
        }
        headers.update(default_headers)
        return web.Response(body=file_sender(file_path=file_path), headers=headers)

    raise web.HTTPForbidden(text="Unsupported file type: " + file_info["type"])


async def on_startup(app):
    """Run on server start."""
    engine = app[ENGINE]
    logger = engine.logger
    logger.info(
        "ImJoy Python Plugin Engine (version %s, api_version %s)",
        __version__,
        API_VERSION,
    )

    if engine.opt.serve:
        logger.info(
            "You can access your local ImJoy web app through %s , imjoy!",
            engine.opt.base_url,
        )
    else:
        logger.info(
            "Please go to https://imjoy.io/#/app "
            "with your web browser (Chrome or FireFox)"
        )
    print("========>> Connection token: {} <<========".format(engine.opt.token))
    sys.stdout.flush()


async def on_shutdown(app):
    """Run on server shut down."""
    engine = app[ENGINE]
    logger = engine.logger
    logger.info("Shutting down the plugin engine")
    stopped = threading.Event()

    def loop():  # executed in another thread
        for i in range(5):
            logger.info("Exiting: %s", 5 - i)
            time.sleep(0.5)
            if stopped.is_set():
                break
        logger.debug("Plugin engine is killed")
        kill_process(os.getpid(), logger)

    loop_thread = threading.Thread(target=loop)
    loop_thread.daemon = True  # stop if the program exits
    loop_thread.start()

    # stopped.set()  # TODO: Should we uncomment this?
    logger.info("Plugin engine exited")
    pid_file = os.path.join(engine.opt.workspace_dir, ".pid")
    try:
        os.remove(pid_file)
    except Exception:  # pylint: disable=broad-except
        logger.info("Failed to remove the pid file")



"""Provide setup function to prepare the engine."""
import argparse
import os
from imjoy import __version__


def parse_cmd_line(args=None):
    """Parse the command line options."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--jupyter", action="store_true", help="start jupyter notebook server"
    )
    parser.add_argument(
        "--legacy", action="store_true", help="start the legacy ImJoy Engine"
    )
    parser.add_argument("--token", type=str, default=None, help="connection token")
    parser.add_argument(
        "--random-token", action="store_true", help="randomly generate a token"
    )
    parser.add_argument("--debug", action="store_true", help="debug mode")
    parser.add_argument(
        "--serve",
        action="store_true",
        help="download ImJoy web app and serve it locally",
    )
    parser.add_argument("--host", type=str, default="127.0.0.1", help="socketio host")
    parser.add_argument(
        "--base_url",
        type=str,
        default=None,
        help="the base url for accessing this plugin engine",
    )
    parser.add_argument("--port", type=str, default="9527", help="socketio port")
    parser.add_argument(
        "--force_quit_timeout",
        type=int,
        default=5,
        help=(
            "the time (in seconds) for waiting before killing a plugin process, "
            "default: 5 s"
        ),
    )
    parser.add_argument(
        "--workspace",
        type=str,
        default=os.path.abspath(os.getcwd()),
        help="workspace folder for plugins",
    )
    parser.add_argument(
        "--freeze", action="store_true", help="disable conda and pip commands"
    )
    parser.add_argument(
        "--engine_container_token",
        type=str,
        default=None,
        help="A token set by the engine container which launches the engine",
    )
    parser.add_argument(
        "--dev", action="store_true", help="run ImJoy Engine in development mode"
    )

    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )

    opt = parser.parse_args(args=args)

    if opt.base_url is None or opt.base_url == "":
        opt.base_url = "http://{}:{}".format(opt.host, opt.port)

    if opt.base_url.endswith("/"):
        opt.base_url = opt.base_url[:-1]

    return opt



"""Provide a mock client."""
import asyncio
import logging
import sys
import uuid
from pathlib import Path

import socketio

from tests.mock_plugin import TestPlugin

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)


class TestClient:
    """Represent a mock client."""

    def __init__(self, url, client_id, session_id, token, loop=None):
        """Set up client instance."""
        self.engine_info = None
        self.sio = None
        self.url = url
        self.client_id = client_id
        self.session_id = session_id
        self.token = token
        self.loop = loop or asyncio.get_event_loop()
        self.plugins = []

    def __repr__(self):
        """Return the client representation."""
        return (
            "<TestClient("
            f"url={self.url}, client_id={self.client_id}, "
            f"session_id={self.session_id}, token={self.token})>"
        )

    async def emit(self, channel, data):
        """Emit a message."""
        fut = self.loop.create_future()

        def callback(ret=None):
            fut.set_result(ret)

        await self.sio.emit(channel, data, callback=callback)
        return await fut

    async def init_plugin(self, plugin_config):
        """Initialize the plugin."""
        pid = plugin_config["name"] + "_" + str(uuid.uuid4())
        ret = await self.emit("init_plugin", {"id": pid, "config": plugin_config})
        assert ret["success"] is True
        secret = ret["secret"]
        plugin = TestPlugin(self.loop, self.sio, pid, secret)
        await plugin.init()
        self.plugins.append(plugin)
        return plugin

    async def connect(self):
        """Connect to the server."""
        sio = socketio.AsyncClient()
        self.sio = sio
        fut = self.loop.create_future()

        @sio.on("connect")
        async def on_connect():  # pylint:disable=unused-variable
            fut.set_result(True)

        @sio.on("disconnect")
        async def on_disconnect():  # pylint:disable=unused-variable
            print("Client disconnected.")
            for plugin in self.plugins:
                plugin.conn.abort.set()

        await sio.connect(self.url)
        return await fut

    async def register_client(self):
        """Register the client."""
        ret = await self.emit(
            "register_client",
            {
                "id": self.client_id,
                "token": self.token,
                "base_url": self.url,
                "session_id": self.session_id,
            },
        )
        if "success" in ret and ret["success"]:
            self.engine_info = ret["engine_info"]
        else:
            logger.error("Failed to register")
            raise Exception("Failed to register")


def main():
    """Run main."""
    home = Path.home()
    workspace_dir = home / "ImJoyWorkspace"
    token_file = workspace_dir / ".token"
    token = token_file.read_text()

    url = "http://localhost:9527"

    client_id = str(uuid.uuid4())
    session_id = str(uuid.uuid4())

    test_plugin_config = {
        "name": "test-plugin",
        "type": "native-python",
        "version": "0.1.12",
        "api_version": "0.1.2",
        "description": "This is a test plugin.",
        "tags": ["CPU", "GPU", "macOS CPU"],
        "ui": "",
        "inputs": None,
        "outputs": None,
        "flags": [],
        "icon": None,
        "env": "conda create -n test-env python=3.6.7",
        "requirements": "pip: numpy",
        "dependencies": [],
    }
    loop = asyncio.get_event_loop()
    client = TestClient(url, client_id, session_id, token, loop)

    async def run():
        await client.connect()
        await client.register_client()
        plugin = await client.init_plugin(test_plugin_config)

    loop.run_until_complete(run())


if __name__ == "__main__":
    main()



"""Provide a mock plugin."""
import asyncio
import logging
import sys

import janus
import numpy as np

from imjoy.utils import dotdict
from imjoy.workers.python_worker import PluginConnection
from imjoy.workers.python3_client import JOB_HANDLERS_PY3

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)

NAME_SPACE = "/"


class ImJoyAPI:
    """ Represent a set of mock ImJoy API """

    def log(self, message):
        logger.info("log: %s", message)

    def error(self, message):
        logger.info("error: %s", message)

    def alert(self, message):
        logger.info("alert: %s", message)

    def showStatus(self, message):
        logger.info("showStatus: %s", message)

    def showMessage(self, message):
        logger.info("showMessage: %s", message)


class TestPlugin:
    """ Represent a mock proxy plugin """

    def __init__(self, loop, sio, pid, secret):
        self.conn = None
        self.loop = loop
        self.sio = sio
        self.pid = pid
        self.secret = secret
        self._plugin_message_handler = []
        self.api = None
        self.imjoy_api = ImJoyAPI()
        self.janus_queue = janus.Queue(loop=self.loop)
        self.queue = self.janus_queue.sync_q

        @sio.on("message_from_plugin_" + secret)
        async def on_message(msg):  # pylint:disable=unused-variable
            logger.info("Message from plugin: %s", msg)
            self.message_handler(msg)

    def get_api(self):
        """return the plugin api functions."""
        return self.conn.local["api"]

    async def message_worker(self, async_q, abort=None):
        """Implement a message worker."""
        while True:
            try:
                if abort is not None and abort.is_set():
                    break

                job = await async_q.get()
                async_q.task_done()
                if job is None:
                    continue

                if "setInterface" == job["type"]:
                    api = self.conn.set_remote(job["api"])
                    self.conn.local["np"] = np
                    self.conn.emit({"type": "interfaceSetAsRemote"})
                    if not self.conn.init:
                        self.conn.set_interface(self.imjoy_api)
                        self.conn.init = True
                    async_q.task_done()
                else:
                    handler = JOB_HANDLERS_PY3.get(job["type"])
                    if handler is None:
                        continue
                    try:
                        await handler(self.conn, job, logger)
                    except Exception:  # pylint: disable=broad-except
                        logger.error(
                            "Error occured in the loop %s", traceback.format_exc()
                        )
                    finally:
                        sys.stdout.flush()
            except Exception as e:
                print(e)

    def terminate(self, msg):
        """mark the plugin as terminated."""
        logger.info("Plugin disconnected: %s", msg)
        self.terminated = True

    def setup(self, conn):

        conn.emit = self.emit

    async def init(self):
        """initialize the plugin."""
        opt = dotdict(id=self.pid, secret=self.secret)
        self.conn = PluginConnection(self, opt)
        self.conn.setup()
        self.terminated = False
        initialized = self.loop.create_future()
        self.on_plugin_message("initialized", initialized)
        self.on_plugin_message("disconnected", self.terminate)
        await initialized

        workers = [
            self.message_worker(self.janus_queue.async_q, self.conn.abort)
            for i in range(2)
        ]
        asyncio.ensure_future(asyncio.gather(*workers))

    async def _emit(self, channel, data):
        """Emit a message."""
        fut = self.loop.create_future()

        def callback(ret=None):
            fut.set_result(ret)

        await self.sio.emit(channel, data, namespace=NAME_SPACE, callback=callback)
        return await fut

    async def emit_plugin_message(self, data):
        """Emit plugin message."""
        await self._emit(
            "message_to_plugin_" + self.secret, {"type": "message", "data": data}
        )

    def emit(self, data):
        """Emit plugin message."""
        asyncio.ensure_future(
            self.emit_plugin_message({"type": "message", "data": data})
        )

    def on_plugin_message(self, message_type, callback_or_future):
        """Add a new plugin message."""
        self._plugin_message_handler.append(
            {"type": message_type, "callback_or_future": callback_or_future}
        )

    async def execute(self, code):
        """Execute plugin code."""
        future = self.loop.create_future()

        def resolve(ret):
            future.set_result(ret)

        def reject(_):
            future.set_exception(Exception("executeFailure"))

        self.on_plugin_message("executeSuccess", resolve)
        self.on_plugin_message("executeFailure", reject)
        await self.emit_plugin_message({"type": "execute", "code": code})
        result = await future
        assert result == {"type": "executeSuccess"}
        await self.emit_plugin_message({"type": "getInterface"})

    def message_handler(self, msg):
        """Handle plugin message."""
        msg_type = msg["type"]
        handlers = self._plugin_message_handler
        for h in handlers:
            # extract message
            if msg_type == "message":
                job = msg["data"]
                self.queue.put(job)
                logger.debug("Added task to the queue")

            elif msg_type == h["type"]:
                callback_or_future = h["callback_or_future"]
                if isinstance(callback_or_future, asyncio.Future):
                    callback_or_future.set_result(msg)
                else:
                    callback_or_future(msg)



import shlex
import shutil
import subprocess
from pathlib import Path

import yaml


def parse_requirements(reqs, conda=False):
    """Parse requirements.

    Return a list of commands to install the requirements.
    """
    commands = []
    if not isinstance(reqs, list):
        reqs = [reqs]
    reqs = [str(req) for req in reqs]

    for req in reqs:
        if ":" in req:
            req_parts = req.split(":")
            typ, libs = req_parts[0], ":".join(req_parts[1:])
            typ, libs = typ.strip(), libs.strip()
            libs = [l.strip() for l in libs.split(" ") if l.strip()]
            if typ == "conda" and libs and conda:
                commands.append("conda install -y " + " ".join(libs))
            elif typ == "pip" and libs:
                commands.append("pip install " + " ".join(libs))
            elif typ == "repo" and libs:
                pass
            elif typ == "cmd" and libs:
                commands.append(" ".join(libs))
            elif "+" in typ or "http" in typ:
                commands.append(f"pip install {req}")
            else:
                raise Exception(f"Unsupported requirement type: {typ}")
        else:
            commands.append(f"pip install {req}")

    return commands


def apply_conda_activate(reqs_cmds, conda_activate, venv_name):
    """Apply the conda activate command to conda install commands."""
    return [conda_activate.format(f"{venv_name} && {cmd}") for cmd in reqs_cmds]


def install_reqs(
    engine, env, work_dir, reqs_cmds, process_start, process_finish, logging_callback
):
    """Install requirements including fallback handling."""
    logger = engine.logger
    cmd_history = engine.store.cmd_history
    commands = []
    code = 0
    errors = None

    for cmd in reqs_cmds:
        if cmd in cmd_history:
            logger.debug("skip command: %s", cmd)
            continue
        commands.append(cmd)

    def fail_install():
        """Notify of failed installation."""
        logging_callback(
            f"Failed to install dependencies with exit code {code} and error: {errors}",
            type="error",
        )
        raise Exception(
            f"Failed to install dependencies with exit code {code} and error: {errors}"
        )

    def _process_start(pid=None, cmd=None):
        """Run after starting the process."""
        logging_callback(f"Running requirements subprocess(pid={pid}): {cmd}")
        process_start(pid=pid)

    logger.info("Running requirements commands: %s", commands)
    code, errors = run_commands(env, work_dir, commands, _process_start, process_finish)

    if code == 0:
        cmd_history.extend(commands)
        logging_callback("Requirements command executed successfully.")
        return
    logging_callback(f"Failed to run requirements command: {commands}", type="error")
    if errors is not None:
        logging_callback(str(errors, "utf-8"), type="error")

    if not engine.opt.conda_available:
        fail_install()

    git_cmd = ""
    if shutil.which("git") is None:
        git_cmd += " git"
    if shutil.which("pip") is None:
        git_cmd += " pip"
    if git_cmd == "":
        fail_install()

    logger.info("pip command failed, trying to install git and pip")
    # try to install git and pip
    git_cmd = f"conda install -y{git_cmd}"
    code, _ = run_process(
        git_cmd.split(),
        process_start=_process_start,
        stderr=None,
        env=env,
        cwd=work_dir,
    )
    if code != 0:
        fail_install()

    logger.info("Running requirements commands: %s", commands)
    code, errors = run_commands(env, work_dir, commands, _process_start, process_finish)

    if code != 0:
        fail_install()


def run_commands(env, work_dir, commands, process_start, process_finish):
    """Run commands in separate processes."""
    code = 0
    errors = None

    for cmd in commands:
        print("Running command:", cmd)
        code, errors = run_process(
            cmd,
            process_start=process_start,
            process_finish=process_finish,
            shell=True,
            env=env,
            cwd=work_dir,
        )
        if code:
            return code, errors

    return code, errors


def run_process(cmd, process_start=None, process_finish=None, **kwargs):
    """Run subprocess command."""
    shell = kwargs.pop("shell", False)
    stderr = kwargs.pop("stderr", subprocess.PIPE)
    process = subprocess.Popen(cmd, shell=shell, stderr=stderr, **kwargs)
    if process_start is not None:
        process_start(pid=process.pid, cmd=cmd)
    return_code = process.wait()
    if stderr is not None:
        _, errors = process.communicate()
    else:
        errors = None
    if process_finish is not None:
        process_finish(pid=process.pid, cmd=cmd)

    return return_code, errors


def parse_env(engine, envs, work_dir, default_env_name):
    """Parse environment."""
    venv_name = None
    is_py2 = False
    logger = engine.logger
    opt = engine.opt

    if isinstance(envs, str):
        envs = envs.strip()

    if not envs:
        if opt.freeze or not opt.conda_available:
            logger.warning(
                "env command is blocked because conda is not available "
                "or in `--freeze` mode: %s",
                envs,
            )
        return venv_name, [], is_py2

    if not isinstance(envs, list):
        envs = [envs]
    for i, env in enumerate(envs):
        if "conda create" in env:
            if "python=2" in env:
                is_py2 = True
            parms = shlex.split(env)
            if "-n" in parms:
                venv_name = parms[parms.index("-n") + 1]
            elif "--name" in parms:
                venv_name = parms[parms.index("--name") + 1]
            else:
                venv_name = default_env_name
                envs[i] = env.replace("conda create", "conda create -n " + venv_name)

            if "-y" not in parms:
                envs[i] = env.replace("conda create", "conda create -y")

        if "conda env create" in env:
            parms = shlex.split(env)
            if "-f" in parms:
                try:
                    env_file = Path(work_dir) / parms[parms.index("-f") + 1]
                    with open(env_file, "r") as stream:
                        env_config = yaml.load(stream)
                        assert "name" in env_config
                        venv_name = env_config["name"]
                except Exception as exc:
                    raise Exception(
                        f"Failed to read env name from the specified env file: {exc}"
                    )

            else:
                raise Exception(
                    "You should provide an environment file via `conda env create -f`"
                )

    if not venv_name or not venv_name.strip():
        venv_name = None

    return venv_name, envs, is_py2



"""Provide a worker template."""
import argparse
import inspect
import logging
import math
import os
import sys
import threading
from functools import reduce
from types import ModuleType

from .utils import ReferenceStore, debounce, dotdict, get_psutil, set_interval
from .python_client import BaseClient

if sys.version_info >= (3, 4):
    from .utils3 import FuturePromise
    from .python3_client import AsyncClient

    PYTHON34 = True
else:
    from .utils import Promise
    from .python_client import Client

    PYTHON34 = False

ARRAY_CHUNK = 1000000
logger = logging.getLogger("plugin")


def kill(proc_pid):
    """Kill process id."""
    psutil = get_psutil()
    if psutil is None:
        return

    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


def ndarray(typed_array, shape, dtype):
    """Return a ndarray."""
    _dtype = type(typed_array)
    if dtype and dtype != _dtype:
        raise Exception(
            "dtype doesn't match the type of the array: " + _dtype + " != " + dtype
        )
    shape = shape or (len(typed_array),)
    return {
        "__jailed_type__": "ndarray",
        "__value__": typed_array,
        "__shape__": shape,
        "__dtype__": _dtype,
    }


API_UTILS = dotdict(
    ndarray=ndarray, kill=kill, debounce=debounce, set_interval=set_interval
)


class PluginConnection:
    """Represent a plugin connection."""

    # pylint:disable=too-many-instance-attributes
    _registered_plugins = {}

    @staticmethod
    def get_plugin(plugin_id):
        return PluginConnection._registered_plugins.get(plugin_id)

    @staticmethod
    def add_plugin(plugin_id, client_id):
        opt = dotdict(id=plugin_id, secret="", work_dir="")
        client = BaseClient.get_client(client_id)
        p = PluginConnection(client, opt)
        return p

    def __init__(self, client, opt):
        """Set up connection instance."""
        self.secret = opt.secret
        self.id = opt.id  # pylint: disable=invalid-name
        self.local = {}
        self._remote = dotdict()
        self.interface = {}
        self.plugin_interfaces = {}
        self.remote_set = False
        self.store = ReferenceStore()
        self.executed = False
        self.init = False
        self.abort = threading.Event()
        self.work_dir = opt.work_dir
        self.opt = opt
        self._registered_plugins[self.id] = self
        self.client = client

        def emit(_):
            raise NotImplementedError

        self.emit = emit

    def setup(self):
        """Set up the plugin connection."""
        assert self.client is not None
        self.client.setup(self)
        if not self.work_dir or self.work_dir == ".":
            self.work_dir = os.getcwd()
        else:
            if not os.path.exists(self.work_dir):
                os.makedirs(self.work_dir)
            os.chdir(self.work_dir)

        self._set_local_api(self._remote)

    def start(self):
        """Start the plugin connection."""
        self.setup()
        self.client.run_forever(self)

    def default_exit(self):
        """Exit default."""
        logger.info("Terminating plugin: %s", self.id)
        self.abort.set()
        os._exit(0)  # pylint: disable=protected-access

    def exit(self, code):
        """Exit."""
        if "exit" in self.interface:
            try:
                self.interface["exit"]()
            except Exception as exc:  # pylint: disable=broad-except
                logger.error("Error when exiting: %s", exc)
                sys.exit(1)
            else:
                logger.info("Terminating plugin")
                sys.exit(code)
        else:
            sys.exit(0)

    def _encode(self, a_object):
        """Encode object."""
        if a_object is None:
            return a_object
        if isinstance(a_object, tuple):
            a_object = list(a_object)
        isarray = isinstance(a_object, list)
        b_object = [] if isarray else {}
        # skip if already encoded
        if (
            isinstance(a_object, dict)
            and "__jailed_type__" in a_object
            and "__value__" in a_object
        ):
            return a_object

        # encode interfaces
        if (
            isinstance(a_object, dict)
            and "__id__" in a_object
            and "__jailed_type__" in a_object
            and a_object["__jailed_type__"] == "plugin_api"
        ):
            encoded_interface = {}
            for key, val in a_object.items():
                if callable(val):
                    b_object[key] = {
                        "__jailed_type__": "plugin_interface",
                        "__plugin_id__": a_object["__id__"],
                        "__value__": key,
                        "num": None,
                    }
                    encoded_interface[key] = val
            self.plugin_interfaces[a_object["__id__"]] = encoded_interface
            return b_object

        keys = range(len(a_object)) if isarray else a_object.keys()
        for key in keys:
            val = a_object[key]
            try:
                basestring
            except NameError:
                basestring = str
            if callable(val):
                interface_func_name = None
                for name in self.interface:
                    if self.interface[name] == val:
                        interface_func_name = name
                        break
                if interface_func_name is None:
                    cid = self.store.put(val)
                    v_obj = {
                        "__jailed_type__": "callback",
                        "__value__": "f",
                        "num": cid,
                    }
                else:
                    v_obj = {
                        "__jailed_type__": "interface",
                        "__value__": interface_func_name,
                    }

            # send objects supported by structure clone algorithm
            # https://developer.mozilla.org/en-US/docs/Web/API/Web_Workers_API/Structured_clone_algorithm
            # if (
            #   v !== Object(v) ||
            #   v instanceof Boolean ||
            #   v instanceof String ||
            #   v instanceof Date ||
            #   v instanceof RegExp ||
            #   v instanceof Blob ||
            #   v instanceof File ||
            #   v instanceof FileList ||
            #   v instanceof ArrayBuffer ||
            #   v instanceof ArrayBufferView ||
            #   v instanceof ImageData
            # ) {
            # }
            elif "np" in self.local and isinstance(
                val, (self.local["np"].ndarray, self.local["np"].generic)
            ):
                v_byte = val.tobytes()
                if len(v_byte) > ARRAY_CHUNK:
                    v_len = int(math.ceil(1.0 * len(v_byte) / ARRAY_CHUNK))
                    v_bytes = []
                    for i in range(v_len):
                        v_bytes.append(v_byte[i * ARRAY_CHUNK : (i + 1) * ARRAY_CHUNK])
                else:
                    v_bytes = v_byte
                v_obj = {
                    "__jailed_type__": "ndarray",
                    "__value__": v_bytes,
                    "__shape__": val.shape,
                    "__dtype__": str(val.dtype),
                }
            elif isinstance(val, (dict, list)):
                v_obj = self._encode(val)
            elif not isinstance(val, basestring) and isinstance(val, bytes):
                v_obj = val.decode()  # covert python3 bytes to str
            elif isinstance(val, Exception):
                v_obj = {"__jailed_type__": "error", "__value__": str(val)}
            else:
                v_obj = {"__jailed_type__": "argument", "__value__": val}

            if isarray:
                b_object.append(v_obj)
            else:
                b_object[key] = v_obj

        return b_object

    def _decode(self, a_object, callback_id, with_promise):
        """Decode object."""
        if a_object is None:
            return a_object
        if "__jailed_type__" in a_object and "__value__" in a_object:
            if a_object["__jailed_type__"] == "callback":
                b_object = self._gen_remote_callback(
                    callback_id, a_object["num"], with_promise
                )
            elif a_object["__jailed_type__"] == "interface":
                name = a_object["__value__"]
                if name in self._remote:
                    b_object = self._remote[name]
                else:
                    b_object = self._gen_remote_method(name)
            elif a_object["__jailed_type__"] == "plugin_interface":
                b_object = self._gen_remote_method(
                    a_object["__value__"], a_object["__plugin_id__"]
                )
            elif a_object["__jailed_type__"] == "ndarray":
                # create build array/tensor if used in the plugin
                try:
                    np = self.local["np"]  # pylint: disable=invalid-name
                    if isinstance(a_object["__value__"], bytes):
                        a_object["__value__"] = a_object["__value__"]
                    elif isinstance(a_object["__value__"], (list, tuple)):
                        a_object["__value__"] = reduce(
                            (lambda x, y: x + y), a_object["__value__"]
                        )
                    else:
                        raise Exception(
                            "Unsupported data type: ",
                            type(a_object["__value__"]),
                            a_object["__value__"],
                        )
                    b_object = np.frombuffer(
                        a_object["__value__"], dtype=a_object["__dtype__"]
                    ).reshape(tuple(a_object["__shape__"]))
                except Exception as exc:
                    logger.debug("Error in converting: %s", exc)
                    b_object = a_object
                    raise exc
            elif a_object["__jailed_type__"] == "error":
                b_object = Exception(a_object["__value__"])
            elif a_object["__jailed_type__"] == "argument":
                b_object = a_object["__value__"]
            else:
                b_object = a_object["__value__"]
            return b_object

        if isinstance(a_object, tuple):
            a_object = list(a_object)
        isarray = isinstance(a_object, list)
        b_object = [] if isarray else dotdict()
        keys = range(len(a_object)) if isarray else a_object.keys()
        for key in keys:
            if isarray or key in a_object:
                val = a_object[key]
                if isinstance(val, (dict, list)):
                    if isarray:
                        b_object.append(self._decode(val, callback_id, with_promise))
                    else:
                        b_object[key] = self._decode(val, callback_id, with_promise)
        return b_object

    def _wrap(self, args):
        """Wrap arguments."""
        wrapped = self._encode(args)
        result = {"args": wrapped}
        return result

    def unwrap(self, args, with_promise):
        """Unwrap arguments."""
        if "callbackId" not in args:
            args["callbackId"] = None
        # wraps each callback so that the only one could be called
        result = self._decode(args["args"], args["callbackId"], with_promise)
        return result

    def set_interface(self, api):
        """Set interface."""
        if isinstance(api, dict):
            api = {a: api[a] for a in api.keys() if not a.startswith("_")}
        elif inspect.isclass(type(api)):
            api = {a: getattr(api, a) for a in dir(api) if not a.startswith("_")}
        else:
            raise Exception("unsupported api export")
        if "exit" in api:
            ext = api["exit"]

            def exit_wrapper():
                try:
                    ext()
                finally:
                    self.default_exit()

            api["exit"] = exit_wrapper
        else:
            api["exit"] = self.default_exit
        self.interface = api
        self.send_interface()

    def send_interface(self):
        """Send interface."""
        names = []
        for name in self.interface:
            if callable(self.interface[name]):
                names.append({"name": name, "data": None})
            else:
                data = self.interface[name]
                if data is not None and isinstance(data, dict):
                    data2 = {}
                    for k in data:
                        if callable(data[k]):
                            data2[k] = "**@@FUNCTION@@**:" + k
                        else:
                            data2[k] = data[k]
                    names.append({"name": name, "data": data2})
                elif isinstance(data, (str, int, float, bool)):
                    names.append({"name": name, "data": data})
        self.emit({"type": "setInterface", "api": names})

    def _gen_remote_method(self, name, plugin_id=None):
        """Return remote method."""

        def remote_method(*arguments, **kwargs):
            """Run remote method."""
            # wrap keywords to a dictionary and pass to the first argument
            if not arguments and kwargs:
                arguments = [kwargs]

            def pfunc(resolve, reject):
                resolve.__jailed_pairs__ = reject
                reject.__jailed_pairs__ = resolve
                call_func = {
                    "type": "method",
                    "name": name,
                    "pid": plugin_id,
                    "args": self._wrap(arguments),
                    "promise": self._wrap([resolve, reject]),
                }
                self.emit(call_func)

            if PYTHON34:
                return FuturePromise(pfunc, self.client.loop)
            return Promise(pfunc)

        remote_method.__remote_method = True  # pylint: disable=protected-access
        return remote_method

    def _gen_remote_callback(self, id_, arg_num, with_promise):
        """Return remote callback."""
        if with_promise:

            def remote_callback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the first argument
                if not arguments and kwargs:
                    arguments = [kwargs]

                def pfunc(resolve, reject):
                    resolve.__jailed_pairs__ = reject
                    reject.__jailed_pairs__ = resolve
                    self.emit(
                        {
                            "type": "callback",
                            "id": id_,
                            "num": arg_num,
                            # 'pid'  : self.id,
                            "args": self._wrap(arguments),
                            "promise": self._wrap([resolve, reject]),
                        }
                    )

                if PYTHON34:
                    return FuturePromise(pfunc, self.client.loop)
                return Promise(pfunc)

        else:

            def remote_callback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the first argument
                if not arguments and kwargs:
                    arguments = [kwargs]
                self.emit(
                    {
                        "type": "callback",
                        "id": id_,
                        "num": arg_num,
                        # 'pid'  : self.id,
                        "args": self._wrap(arguments),
                    }
                )

        return remote_callback

    def set_remote(self, api):
        """Set remote."""
        _remote = dotdict()
        for i, _ in enumerate(api):
            if isinstance(api[i], dict) and "name" in api[i]:
                name = api[i]["name"]
                data = api[i].get("data", None)
                if data is not None:
                    if isinstance(data, dict):
                        data2 = dotdict()
                        for key in data:
                            if key in data:
                                if data[key] == "**@@FUNCTION@@**:" + key:
                                    data2[key] = self._gen_remote_method(
                                        name + "." + key
                                    )
                                else:
                                    data2[key] = data[key]
                        _remote[name] = data2
                    else:
                        _remote[name] = data
                else:
                    _remote[name] = self._gen_remote_method(name)

        self._set_local_api(_remote)
        return _remote

    def _set_local_api(self, _remote):
        """Set local API."""
        _remote["export"] = self.set_interface
        _remote["utils"] = API_UTILS
        _remote["WORK_DIR"] = self.work_dir

        self.local["api"] = _remote

        # make a fake module with api
        mod = ModuleType("imjoy")
        sys.modules[mod.__name__] = mod  # pylint: disable=no-member
        mod.__file__ = mod.__name__ + ".py"  # pylint: disable=no-member
        mod.api = _remote


def main():
    """Run script."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--id", type=str, required=True, help="plugin id")
    parser.add_argument("--secret", type=str, required=True, help="plugin secret")
    parser.add_argument(
        "--work_dir", type=str, default=".", help="plugin working directory"
    )
    parser.add_argument(
        "--server", type=str, default="http://127.0.0.1:9527", help="socketio server"
    )
    parser.add_argument("--daemon", action="store_true", help="daemon mode")
    parser.add_argument("--debug", action="store_true", help="debug mode")

    opt = parser.parse_args()

    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.INFO)
    if opt.debug:
        logger.setLevel(logging.DEBUG)

    if PYTHON34:
        client = AsyncClient()
    else:
        client = Client()
    plugin_conn = PluginConnection(client, opt)
    plugin_conn.start()


if __name__ == "__main__":
    main()



"""Provide main entrypoint."""
import json
import os
import subprocess
import sys

from imjoy.options import parse_cmd_line
from imjoy.utils import read_or_generate_token, write_token


def main():
    """Run main."""
    opt = parse_cmd_line()
    if opt.jupyter:
        sys.argv = sys.argv[:1]
        sys.argc = 1
        from notebook.notebookapp import NotebookApp

        kwargs = {
            "open_browser": False,
            "allow_origin": "*",
            "ip": opt.host,
            "notebook_dir": opt.workspace,
            "port": int(opt.port),
            "tornado_settings": {"headers": {"Access-Control-Allow-Origin": "*"}},
        }

        if opt.token is None or opt.token == "":
            if not opt.random_token:
                opt.token = read_or_generate_token()
                kwargs["token"] = opt.token
        else:
            kwargs["token"] = opt.token

        app = NotebookApp.instance(**kwargs)
        app.initialize()
        if app.port != int(opt.port):
            print("\nWARNING: using a different port: {}.\n".format(app.port))
        write_token(app.token)
        app._token_generated = True
        app.start()
        return
    if not opt.legacy:
        print(
            "\nNote: We are migrating the backend of the ImJoy Engine to Jupyter, to use it please run `imjoy --jupyter`.\n\nIf you want to use the previous engine, run `imjoy --legacy`, however, please note that it maybe removed soon.\n"
        )
        return

    if opt.dev:
        print("Running ImJoy Plugin Engine in development mode")
        from .engine import run

        run()
        return

    # add executable path to PATH
    os.environ["PATH"] = (
        os.path.split(sys.executable)[0] + os.pathsep + os.environ.get("PATH", "")
    )
    conda_available = False
    try:
        # for fixing CondaHTTPError:
        # https://github.com/conda/conda/issues/6064#issuecomment-458389796
        process = subprocess.Popen(
            ["conda", "info", "--json", "-s"], stdout=subprocess.PIPE
        )
        cout, _ = process.communicate()
        conda_prefix = json.loads(cout.decode("ascii"))["conda_prefix"]
        print("Found conda environment: {}".format(conda_prefix))
        conda_available = True
        if os.name == "nt":
            os.environ["PATH"] = (
                os.path.join(conda_prefix, "Library", "bin")
                + os.pathsep
                + os.environ["PATH"]
            )
    except OSError:
        if sys.version_info > (3, 0):
            print(
                "WARNING: you are running ImJoy without conda, "
                "you may have problem with some plugins"
            )
        conda_prefix = None

    pip_available = True
    try:
        import distutils.spawn

        if distutils.spawn.find_executable("git") is None:
            if not opt.freeze and conda_available:
                print("git not found, trying to install with conda")
                # try to install git
                ret = subprocess.Popen(
                    "conda install -y git".split(), shell=False
                ).wait()
                if ret != 0:
                    raise Exception(
                        "Failed to install git/pip and dependencies "
                        "with exit code: {}".format(ret)
                    )
            elif not opt.freeze:
                print(
                    "git not found, unable to install it because conda is not available"
                )

        if distutils.spawn.find_executable("pip") is None:
            pip_available = False
            if not opt.freeze and conda_available:
                print("pip not found, trying to install pip with conda")
                # try to install pip
                ret = subprocess.Popen(
                    "conda install -y pip".split(), shell=False
                ).wait()
                if ret != 0:
                    raise Exception(
                        "Failed to install git/pip and dependencies "
                        "with exit code: {}".format(ret)
                    )
                else:
                    pip_available = True
            elif not opt.freeze:
                print(
                    "pip not found, unable to install it because conda is not available"
                )
        else:
            print("Upgrading pip...")
            ret = subprocess.Popen(
                "python -m pip install -U pip".split(), shell=False
            ).wait()
            if ret != 0:
                print("Failed to upgrade pip.")
    except Exception as e:
        print("Failed to check or install pip/git, error: {}".format(e))

    try:
        import psutil
    except ImportError:
        if not opt.freeze:
            if pip_available:
                print("Trying to install psutil with pip")
                ret = subprocess.Popen(
                    "python -m pip install psutil".split(),
                    env=os.environ.copy(),
                    shell=False,
                ).wait()
            else:
                ret = 1

            if ret != 0 and conda_available:
                print("Trying to install psutil with conda")
                ret2 = subprocess.Popen(
                    "conda install -y psutil".split(), env=os.environ.copy()
                ).wait()
                if ret2 != 0:
                    print(
                        "WARNING: Failed to install psutil, "
                        "please try to setup an environment with gcc support."
                    )
                else:
                    print("psutil was installed successfully.")
            elif ret != 0:
                print("Failed to install psutil.")

    if sys.version_info > (3, 0):
        if not opt.freeze and pip_available:
            # running in python 3
            print("Upgrading ImJoy Plugin Engine")
            ret = subprocess.Popen(
                "python -m pip install -U imjoy[engine]".split(),
                env=os.environ.copy(),
                shell=False,
            ).wait()
            if ret != 0:
                print("Failed to upgrade ImJoy Plugin Engine")
        elif not opt.freeze:
            print("Failed to upgrade the engine because pip was not found.")

        # reload to use the new version
        import imjoy
        import importlib

        importlib.reload(imjoy)
        from imjoy.engine import run

        run()
    else:
        # running in python 2
        if conda_available:
            print("ImJoy needs to run in Python 3.6+, bootstrapping with conda")
            ret = subprocess.Popen(
                "conda create -y -n imjoy python=3.6 conda".split(),
                env=os.environ.copy(),
                shell=False,
            ).wait()
            if ret == 0:
                print(
                    "conda environment is now ready, "
                    "installing imjoy and starting the engine"
                )
            else:
                print(
                    "conda environment failed to setup, maybe it already exists. "
                    "Otherwise, please make sure you are running in a conda environment"
                )
            pip_cmd = "python -m pip install -U imjoy[engine]"

            if sys.platform == "linux" or sys.platform == "linux2":
                # linux
                conda_activate = (
                    "/bin/bash -c 'source " + conda_prefix + "/bin/activate {}'"
                )
            elif sys.platform == "darwin":
                # OS X
                conda_activate = "source activate {}"
            elif sys.platform == "win32":
                # Windows...
                conda_activate = "activate {}"
            else:
                conda_activate = "conda activate {}"

            pip_cmd = conda_activate.format(
                "imjoy && " + pip_cmd + " && python -m imjoy"
            )
            print("Running command: " + pip_cmd)
            ret = subprocess.Popen(pip_cmd, shell=True).wait()
            if ret != 0:
                raise Exception(
                    "Failed to install and start ImJoy, exit code: {}".format(ret)
                )
        else:
            raise Exception(
                "It seems you are trying to run ImJoy Engine in Python 2, but it requires Python 3.6+ (with conda)."
            )


if __name__ == "__main__":
    main()



"""Provide a plugin manager."""
import asyncio
import datetime
import os
import platform
import subprocess
import sys
import threading
import time
import traceback
import uuid

import GPUtil

from imjoy import __version__, API_VERSION, IMJOY_PACKAGE_DIR
from imjoy.connection.decorator import socketio_handler as sio_on
from imjoy.utils import console_to_str, parse_repos, get_psutil, kill_process
from .helper import (
    apply_conda_activate,
    install_reqs,
    parse_env,
    parse_requirements,
    run_commands,
    run_process,
)

MAX_ATTEMPTS = 1000


def setup_subprocess_runner(engine):
    """Set up the subprocess runner."""
    engine.conn.register_event_handler(on_register_client)
    engine.conn.register_event_handler(on_init_plugin)
    engine.conn.register_event_handler(on_kill_plugin)
    engine.conn.register_event_handler(on_kill_plugin_process)
    engine.conn.register_event_handler(disconnect_client_session)
    engine.conn.register_event_handler(disconnect_plugin)
    engine.conn.register_event_handler(reset_engine_plugins)


@sio_on("init_plugin")
async def on_init_plugin(engine, sid, kwargs):
    """Initialize plugin."""
    logger = engine.logger
    registered_sessions = engine.store.registered_sessions
    try:
        if sid in registered_sessions:
            obj = registered_sessions[sid]
            client_id, session_id = obj["client"], obj["session"]
        else:
            logger.debug("Client %s is not registered", sid)
            return {"success": False}
        pid = kwargs["id"]
        config = kwargs.get("config", {})
        env = config.get("env")
        cmd = config.get("cmd", "python")
        pname = config.get("name")
        flags = config.get("flags", [])
        tag = config.get("tag", "")
        requirements = config.get("requirements", []) or []
        workspace = config.get("workspace", "default")
        work_dir = os.path.join(engine.opt.workspace_dir, workspace)
        if not os.path.exists(work_dir):
            os.makedirs(work_dir)
        plugin_env = os.environ.copy()
        plugin_env["WORK_DIR"] = work_dir
        if engine.opt.dev:
            plugin_env["PYTHONPATH"] = (
                IMJOY_PACKAGE_DIR + os.path.pathsep + plugin_env.get("PYTHONPATH", "")
            )
            worker_module = "workers.python_worker"
            logger.debug(
                "ImJoy package directory was added to PYTHONPATH, will run module `workers.python_worker`."
            )
        else:
            worker_module = "imjoy.workers.python_worker"
            logger.debug(
                "Will run module `imjoy.workers.python_worker` from installed ImJoy package."
            )

        logger.info(
            "Initialize the plugin, name=%s, id=%s, cmd=%s, workspace=%s",
            pname,
            pid,
            cmd,
            workspace,
        )

        if "single-instance" in flags:
            plugin_signature = "{}/{}".format(pname, tag)
            resume = True
        elif "allow-detach" in flags:
            plugin_signature = "{}/{}/{}/{}".format(client_id, workspace, pname, tag)
            resume = True
        else:
            plugin_signature = None
            resume = False

        if resume:
            plugin_info = resume_plugin_session(
                engine, pid, session_id, plugin_signature
            )
            if plugin_info is not None:
                if "aborting" in plugin_info:
                    logger.info("Waiting for plugin %s to abort", plugin_info["id"])
                    await plugin_info["aborting"]
                else:
                    logger.debug("Plugin already initialized: %s", pid)
                    return {
                        "success": True,
                        "resumed": True,
                        "initialized": True,
                        "secret": plugin_info["secret"],
                        "work_dir": os.path.abspath(work_dir),
                    }
            else:
                logger.info(
                    "Failed to resume single instance plugin: %s, %s",
                    pid,
                    plugin_signature,
                )

        secret_key = str(uuid.uuid4())
        abort = threading.Event()
        plugin_info = {
            "secret": secret_key,
            "id": pid,
            "abort": abort,
            "flags": flags,
            "session_id": session_id,
            "name": config["name"],
            "type": config["type"],
            "client_id": client_id,
            "signature": plugin_signature,
            "process_id": None,
        }
        logger.debug("Add plugin: %s", plugin_info)
        add_plugin(engine, plugin_info)

        @sio_on("from_plugin_" + secret_key)
        async def message_from_plugin(engine, sid, kwargs):
            if kwargs["type"] in [
                "initialized",
                "importSuccess",
                "importFailure",
                "executeSuccess",
                "executeFailure",
            ]:
                await engine.conn.sio.emit("message_from_plugin_" + secret_key, kwargs)
                logger.debug("Message from %s", pid)
                if kwargs["type"] == "initialized":
                    add_plugin(engine, plugin_info, sid)
                elif kwargs["type"] == "executeFailure":
                    logger.info("Killing plugin %s due to exeuction failure", pid)
                    kill_plugin(engine, pid)
            else:
                await engine.conn.sio.emit(
                    "message_from_plugin_" + secret_key,
                    {"type": "message", "data": kwargs},
                )

        engine.conn.register_event_handler(message_from_plugin)

        @sio_on("message_to_plugin_" + secret_key)
        async def message_to_plugin(engine, _, kwargs):
            if kwargs["type"] == "message":
                await engine.conn.sio.emit("to_plugin_" + secret_key, kwargs["data"])
            logger.debug("Message to plugin %s, %s", secret_key, kwargs["data"])

        engine.conn.register_event_handler(message_to_plugin)

        eloop = asyncio.get_event_loop()

        def stop_callback(success, message):
            if "aborting" in plugin_info:
                plugin_info["aborting"].set_result(success)
            message = str(message or "")
            logger.info(
                "Disconnecting from plugin (success: %s, message: %s)", success, message
            )
            coro = engine.conn.sio.emit(
                "message_from_plugin_" + secret_key,
                {
                    "type": "disconnected",
                    "details": {"success": success, "message": message},
                },
            )
            asyncio.run_coroutine_threadsafe(coro, eloop).result()

        def logging_callback(msg, type="info"):  # pylint: disable=redefined-builtin
            if msg == "":
                return
            coro = engine.conn.sio.emit(
                "message_from_plugin_" + secret_key,
                {"type": "logging", "details": {"value": msg, "type": type}},
            )
            asyncio.run_coroutine_threadsafe(coro, eloop).result()

        args = '{} -m {} --id="{}" --server={} --secret="{}"'.format(
            cmd, worker_module, pid, "http://127.0.0.1:" + engine.opt.port, secret_key
        )
        task_thread = threading.Thread(
            target=launch_plugin,
            args=[
                engine,
                stop_callback,
                logging_callback,
                pid,
                pname,
                tag,
                env,
                requirements,
                args,
                work_dir,
                abort,
                pid,
                plugin_env,
            ],
        )
        task_thread.daemon = True
        task_thread.start()
        return {
            "success": True,
            "initialized": False,
            "secret": secret_key,
            "work_dir": os.path.abspath(work_dir),
        }

    except Exception:  # pylint: disable=broad-except
        traceback_error = traceback.format_exc()
        logger.error(traceback_error)
        return {"success": False, "reason": traceback_error}


@sio_on("register_client")
async def on_register_client(engine, sid, kwargs):
    """Register client."""
    logger = engine.logger
    conn_data = engine.store
    logger.info("Registering client: %s", kwargs)
    client_id = kwargs.get("id", str(uuid.uuid4()))
    workspace = kwargs.get("workspace", "default")
    session_id = kwargs.get("session_id", str(uuid.uuid4()))
    base_url = kwargs.get("base_url", engine.opt.base_url)
    if base_url.endswith("/"):
        base_url = base_url[:-1]

    token = kwargs.get("token")
    if token != engine.opt.token:
        logger.error("Wrong connection token (%s)", token)
        print("========>> Connection token: {} <<========".format(engine.opt.token))
        if engine.opt.engine_container_token is not None:
            await engine.conn.sio.emit(
                "message_to_container_" + engine.opt.engine_container_token,
                {
                    "type": "popup_token",
                    "client_id": client_id,
                    "session_id": session_id,
                },
            )
        # try:
        #     webbrowser.open(
        #         'http://'+opt.host+':'+opt.port+'/about?token='+opt.token,
        #         new=0, autoraise=True)
        # except Exception as exc:
        #     logger.error("Failed to open the browser: %s", exc)
        conn_data.attempt_count += 1
        if conn_data.attempt_count >= MAX_ATTEMPTS:
            logger.info(
                "Client exited because max attemps exceeded: %s",
                conn_data.attempt_count,
            )
            sys.exit(100)
        return {"success": False, "message": "Connection token mismatch."}

    conn_data.attempt_count = 0
    if add_client_session(engine, session_id, client_id, sid, base_url, workspace):
        confirmation = True
        message = (
            "Another ImJoy session is connected to this Plugin Engine({}), "
            "allow a new session to connect?".format(base_url)
        )
    else:
        confirmation = False
        message = None

    engine_info = {"api_version": API_VERSION, "version": __version__}
    engine_info["platform"] = {
        "uname": ", ".join(platform.uname()),
        "machine": platform.machine(),
        "system": platform.system(),
        "processor": platform.processor(),
        "node": platform.node(),
    }

    try:
        gpus = GPUtil.getGPUs()
        engine_info["GPUs"] = [
            {
                "name": gpu.name,
                "id": gpu.id,
                "memory_total": gpu.memoryTotal,
                "memory_util": gpu.memoryUtil,
                "memoryUsed": gpu.memoryUsed,
                "driver": gpu.driver,
                "temperature": gpu.temperature,
                "load": gpu.load,
            }
            for gpu in gpus
        ]
    except Exception:  # pylint: disable=broad-except
        logger.error("Failed to get GPU information with GPUtil")

    return {
        "success": True,
        "confirmation": confirmation,
        "message": message,
        "engine_info": engine_info,
    }


@sio_on("kill_plugin")
async def on_kill_plugin(engine, sid, kwargs):
    """Kill plugin."""
    logger = engine.logger
    plugins = engine.store.plugins
    registered_sessions = engine.store.registered_sessions
    logger.info("Kill plugin: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered"}

    pid = kwargs["id"]
    if pid in plugins:
        if "killing" not in plugins[pid]:
            obj = {"force_kill": True, "pid": pid}
            plugins[pid]["killing"] = True

            def exited(_):
                obj["force_kill"] = False
                logger.info("Plugin %s exited normally", pid)
                # kill the plugin now
                kill_plugin(engine, pid)

            await engine.conn.sio.emit(
                "to_plugin_" + plugins[pid]["secret"],
                {"type": "disconnect"},
                callback=exited,
            )
            await force_kill_timeout(engine, engine.opt.force_quit_timeout, obj)
    return {"success": True}


@sio_on("kill_plugin_process")
async def on_kill_plugin_process(engine, sid, kwargs):
    """Kill plugin process."""
    logger = engine.logger
    plugins = engine.store.plugins
    registered_sessions = engine.store.registered_sessions
    if sid not in registered_sessions:
        logger.debug("Client %s is not registered", sid)
        return {"success": False, "error": "client has not been registered."}
    if "all" not in kwargs:
        return {
            "success": False,
            "error": 'You must provide the pid of the plugin process or "all=true".',
        }
    if kwargs["all"]:
        logger.info("Killing all the plugins")
        await kill_all_plugins(engine, sid)
        return {"success": True}
    try:
        kill_process(int(kwargs["pid"]), logger)
        return {"success": True}
    except Exception:  # pylint: disable=broad-except
        return {
            "success": False,
            "error": "Failed to kill plugin process: #" + str(kwargs["pid"]),
        }

    psutil = get_psutil()
    if not psutil:
        return
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    pids = []
    for proc in children:
        if proc.is_running() and proc.status() != psutil.STATUS_ZOMBIE:
            pids.append(proc.pid)
    # remove plugin if the corresponding process does not exist any more
    for plugin in plugins.values():
        if plugin["process_id"] not in pids:
            plugin["process_id"] = None
            kill_plugin(engine, plugin["id"])


def resume_plugin_session(engine, pid, session_id, plugin_signature):
    """Resume plugin session."""
    logger = engine.logger
    plugins = engine.store.plugins
    plugin_sessions = engine.store.plugin_sessions
    plugin_signatures = engine.store.plugin_signatures
    if pid in plugins:
        if session_id in plugin_sessions:
            plugin_sessions[session_id].append(plugins[pid])
        else:
            plugin_sessions[session_id] = [plugins[pid]]

    if plugin_signature in plugin_signatures:
        plugin_info = plugin_signatures[plugin_signature]
        logger.info("Resuming plugin %s", pid)
        return plugin_info

    return None


def add_client_session(engine, session_id, client_id, sid, base_url, workspace):
    """Add client session."""
    logger = engine.logger
    clients = engine.store.clients
    registered_sessions = engine.store.registered_sessions
    if client_id in clients:
        clients[client_id].append(sid)
        client_connected = True
    else:
        clients[client_id] = [sid]
        client_connected = False
    logger.info("Adding client session %s", sid)
    registered_sessions[sid] = {
        "client": client_id,
        "session": session_id,
        "base_url": base_url,
        "workspace": workspace,
    }
    return client_connected


@sio_on("disconnect_client_session")
async def disconnect_client_session(engine, sid):
    """Disconnect client session."""
    logger = engine.logger
    clients = engine.store.clients
    plugin_sessions = engine.store.plugin_sessions
    registered_sessions = engine.store.registered_sessions
    if sid in registered_sessions:
        logger.info("Disconnecting client session %s", sid)
        obj = registered_sessions[sid]
        client_id, session_id = obj["client"], obj["session"]
        del registered_sessions[sid]
        if client_id in clients and sid in clients[client_id]:
            clients[client_id].remove(sid)
            if not clients[client_id]:
                del clients[client_id]
        if session_id in plugin_sessions:
            for plugin in plugin_sessions[session_id]:
                if "allow-detach" not in plugin["flags"]:
                    kill_plugin(engine, plugin["id"])
            del plugin_sessions[session_id]


def add_plugin(engine, plugin_info, sid=None):
    """Add plugin."""
    plugins = engine.store.plugins
    plugin_sessions = engine.store.plugin_sessions
    plugin_sids = engine.store.plugin_sids
    plugin_signatures = engine.store.plugin_signatures
    pid = plugin_info["id"]
    session_id = plugin_info["session_id"]
    plugin_signatures[plugin_info["signature"]] = plugin_info

    if pid not in plugins:
        if session_id in plugin_sessions:
            plugin_sessions[session_id].append(plugin_info)
        else:
            plugin_sessions[session_id] = [plugin_info]
    plugins[pid] = plugin_info

    if pid in plugins and sid is not None:
        plugin_sids[sid] = plugin_info
        plugin_info["sid"] = sid


@sio_on("disconnect_plugin")
async def disconnect_plugin(engine, sid):
    """Disconnect plugin."""
    logger = engine.logger
    plugins = engine.store.plugins
    plugin_sessions = engine.store.plugin_sessions
    plugin_sids = engine.store.plugin_sids
    plugin_signatures = engine.store.plugin_signatures
    if sid in plugin_sids:
        logger.info("Disconnecting plugin session %s", sid)
        pid = plugin_sids[sid]["id"]
        if pid in plugins:
            logger.info("Cleaning up plugin %s", pid)
            if plugins[pid]["signature"] in plugin_signatures:
                logger.info(
                    "Cleaning up plugin signature %s", plugins[pid]["signature"]
                )
                del plugin_signatures[plugins[pid]["signature"]]
            del plugins[pid]
        del plugin_sids[sid]
        for session_id in plugin_sessions.keys():
            exist = False
            for plugin in plugin_sessions[session_id]:
                if plugin["id"] == pid:
                    exist = plugin
            if exist:
                logger.info("Cleaning up plugin session %s", session_id)
                plugin_sessions[session_id].remove(exist)
                kill_plugin(engine, exist["id"])


def set_plugin_pid(engine, plugin_id, pid):
    """Set plugin pid."""
    plugins = engine.store.plugins
    plugins[plugin_id]["process_id"] = pid


def kill_plugin(engine, pid):
    """Kill plugin."""
    logger = engine.logger
    plugins = engine.store.plugins
    plugin_sids = engine.store.plugin_sids
    plugin_signatures = engine.store.plugin_signatures
    if pid in plugins:
        try:
            plugins[pid]["abort"].set()
            plugins[pid]["aborting"] = asyncio.get_event_loop().create_future()
            if plugins[pid]["process_id"] is not None:
                kill_process(plugins[pid]["process_id"], logger)
        except Exception as exc:  # pylint: disable=broad-except
            logger.error("Failed to kill plugin %s, error: %s", pid, exc)
        if "sid" in plugins[pid]:
            if plugins[pid]["sid"] in plugin_sids:
                del plugin_sids[plugins[pid]["sid"]]

        if plugins[pid]["signature"] in plugin_signatures:
            logger.info(
                "Cleaning up killed plugin signature %s", plugins[pid]["signature"]
            )
            del plugin_signatures[plugins[pid]["signature"]]
        logger.info("Cleaning up killed plugin %s", pid)
        del plugins[pid]


@sio_on("reset_engine_plugins")
async def reset_engine_plugins(engine, sid, _):
    """Handle plugins when reset engine is called."""
    await kill_all_plugins(engine, sid)


async def kill_all_plugins(engine, sid):
    """Kill all plugins."""
    logger = engine.logger
    plugin_sids = engine.store.plugin_sids
    # copy dict as it will change size when killing plugins
    for plugin_info in dict(plugin_sids).values():
        try:
            await on_kill_plugin(engine, sid, {"id": plugin_info["id"]})
        except Exception as exc:  # pylint: disable=broad-except
            logger.error("%s", exc)


async def force_kill_timeout(engine, timeout, obj):
    """Force kill plugin after timeout."""
    logger = engine.logger
    pid = obj["pid"]
    for _ in range(int(timeout * 10)):
        if obj["force_kill"]:
            await asyncio.sleep(0.1)
        else:
            return

    logger.warning("Timeout, force quitting %s", pid)
    try:
        kill_plugin(engine, pid)
    except Exception as exc:  # pylint:disable=broad-except
        logger.error("Failed to kill plugin %s: %s", pid, exc)


def launch_plugin(
    engine,
    stop_callback,
    logging_callback,
    plugin_id,
    pname,
    tag,
    env,
    requirements,
    args,
    work_dir,
    abort,
    name,
    plugin_env,
):
    """Launch plugin."""
    logger = engine.logger
    opt = engine.opt
    if abort.is_set():
        logger.info("Plugin aborting")
        logging_callback("Plugin aborting")
        return False
    venv_name = None
    progress = 0
    try:
        repos = parse_repos(requirements, work_dir)
        progress = 5
        logging_callback(progress, type="progress")
        for repo in repos:
            try:
                logger.info("Cloning repo %s to %s", repo["url"], repo["repo_dir"])
                logging_callback(f"Cloning repo {repo['url']} to {repo['repo_dir']}")
                if os.path.exists(repo["repo_dir"]):
                    assert os.path.isdir(repo["repo_dir"])
                    cmd = "git pull --all"
                    run_cmd(
                        engine,
                        cmd.split(" "),
                        cwd=repo["repo_dir"],
                        plugin_id=plugin_id,
                    )
                else:
                    cmd = (
                        "git clone --progress --depth=1 "
                        + repo["url"]
                        + " "
                        + repo["repo_dir"]
                    )
                    run_cmd(engine, cmd.split(" "), cwd=work_dir, plugin_id=plugin_id)
                progress += int(20 / len(repos))
                logging_callback(progress, type="progress")
            except Exception as exc:  # pylint: disable=broad-except
                logging_callback(f"Failed to obtain the git repo: {exc}", type="error")

        default_virtual_env = "{}-{}".format(pname, tag) if tag != "" else pname
        default_virtual_env = default_virtual_env.replace(" ", "_")
        venv_name, envs, is_py2 = parse_env(engine, env, work_dir, default_virtual_env)
        environment_variables = {}

        if engine.opt.dev:
            default_requirements = ["imjoy[worker]"]
        else:
            default_requirements = ["imjoy[worker]==" + __version__]
        default_reqs_cmds = parse_requirements(
            default_requirements, conda=opt.conda_available
        )
        reqs_cmds = parse_requirements(requirements, opt.conda_available)
        reqs_cmds += default_reqs_cmds

        cmd_history = engine.store.cmd_history

        def process_start(pid=None, cmd=None):
            """Run before process starts."""
            if pid is not None:
                set_plugin_pid(engine, plugin_id, pid)
            if cmd is not None:
                logger.info("Running command %s", cmd)

        def process_finish(pid=None, cmd=None):
            """Notify when an install process command has finished."""
            logger.debug("Finished running (pid=%s): %s", pid, cmd)
            nonlocal progress
            progress += int(70 / (len(envs) + len(reqs_cmds)))
            logging_callback(progress, type="progress")

        for _env in envs:
            if isinstance(_env, str):
                if _env not in cmd_history:
                    logger.info("Running env command: %s", _env)
                    logging_callback(f"Running env command: {_env}")
                    code, errors = run_process(
                        _env.split(),
                        process_start=process_start,
                        process_finish=process_finish,
                        env=plugin_env,
                        cwd=work_dir,
                    )

                    if code == 0:
                        cmd_history.append(_env)
                        logging_callback("Successful execution of env command")

                    if errors is not None:
                        logging_callback(str(errors, "utf-8"), type="error")

                else:
                    logger.debug("Skip env command: %s", _env)
                    logging_callback(f"Skip env command: {_env}")

            elif isinstance(_env, dict):
                assert "type" in _env
                if _env["type"] == "gputil":
                    # Set CUDA_DEVICE_ORDER
                    # so the IDs assigned by CUDA match those from nvidia-smi
                    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
                    device_ids = GPUtil.getAvailable(**_env["options"])
                    if not device_ids:
                        raise Exception("No GPU is available to run this plugin.")
                    environment_variables["CUDA_VISIBLE_DEVICES"] = ",".join(
                        [str(device_id) for device_id in device_ids]
                    )
                    logging_callback(f"GPU id assigned: {device_ids}")
                elif _env["type"] == "variable":
                    environment_variables.update(_env["options"])
            else:
                logger.debug("Skip unsupported env: %s", _env)

            if abort.is_set():
                logger.info("Plugin aborting")
                return False

        if opt.freeze:
            logger.warning(
                "pip command is blocked due to `--freeze` mode: %s", reqs_cmds
            )
            reqs_cmds = []

        elif opt.conda_available and venv_name is not None:
            reqs_cmds = apply_conda_activate(reqs_cmds, opt.conda_activate, venv_name)

        install_reqs(
            engine,
            plugin_env,
            work_dir,
            reqs_cmds,
            process_start,
            process_finish,
            logging_callback,
        )

    except Exception:  # pylint: disable=broad-except
        error_traceback = traceback.format_exc()
        logger.error(
            "Failed to setup plugin virtual environment or its requirements: %s",
            error_traceback,
        )
        abort.set()
        stop_callback(False, f"Plugin process failed to start: {error_traceback}")
        return False

    if abort.is_set():
        logger.info("Plugin aborting")
        stop_callback(False, "Plugin process failed to start")
        return False
    # env = os.environ.copy()
    if opt.conda_available and venv_name is not None:
        [args] = apply_conda_activate([args], opt.conda_activate, venv_name)
    if isinstance(args, str):
        args = args.split()
    if not args:
        args = []
    # Convert them all to strings
    args = [str(x) for x in args if str(x) != ""]
    logger.info("Plugin %s task started", name)

    args = " ".join(args)
    logger.info("Task subprocess args: %s", args)

    # set system/version dependent "start_new_session" analogs
    # https://docs.python.org/2/library/subprocess.html#converting-argument-sequence
    kwargs = {}
    if sys.platform != "win32":
        kwargs.update(preexec_fn=os.setsid)
    progress = 100
    logging_callback(progress, type="progress")
    try:
        _env = plugin_env.copy()
        _env.update(environment_variables)
        logger.debug("Running process with env: %s", _env)
        process = subprocess.Popen(
            args,
            bufsize=1,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            env=_env,
            cwd=work_dir,
            **kwargs,
        )
        logging_callback(f"Running subprocess (pid={process.pid}) with {args}")
        set_plugin_pid(engine, plugin_id, process.pid)
        # Poll process for new output until finished
        stdfn = sys.stdout.fileno()

        progress = 0
        logging_callback(progress, type="progress")

        while True:
            out = process.stdout.read(1)
            if out == "" and process.poll() is not None:
                break
            os.write(stdfn, out)
            sys.stdout.flush()
            if abort.is_set() or process.poll() is not None:
                break
            time.sleep(0)

        logger.info("Plugin aborting")
        kill_process(process.pid, logger)

        outputs, errors = process.communicate()
        if outputs is not None:
            outputs = str(outputs, "utf-8")
        if errors is not None:
            errors = str(errors, "utf-8")
        exit_code = process.returncode
    except Exception as exc:  # pylint: disable=broad-except
        logger.error(traceback.format_exc())
        outputs, errors = "", str(exc)
        exit_code = 100
    if exit_code == 0:
        logging_callback(f"Plugin process exited with code {exit_code}")
        stop_callback(True, outputs)
        return True

    logging_callback(f"Plugin process exited with code {exit_code}", type="error")
    logger.error(
        "Error occured during terminating a process.\n" "Command: %s\nExit code: %s",
        args,
        exit_code,
    )
    errors = errors or ""
    stop_callback(False, f"{errors}\nPlugin process exited with code {exit_code}")
    return False


def run_cmd(
    engine,
    cmd,
    shell=False,
    cwd=None,
    check_returncode=True,
    callback=None,
    plugin_id=None,
):
    """Run command.

    From https://github.com/vcs-python/libvcs/.
    """
    proc = subprocess.Popen(
        cmd,
        shell=shell,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        creationflags=0,
        bufsize=1,
        cwd=cwd,
    )
    if plugin_id is not None:
        set_plugin_pid(engine, plugin_id, proc.pid)

    all_output = []
    code = None
    line = None
    while code is None:
        code = proc.poll()
        if callback and callable(callback):
            line = console_to_str(proc.stderr.read(128))
            if line:
                callback(output=line, timestamp=datetime.datetime.now())
    if callback and callable(callback):
        callback(output="\r", timestamp=datetime.datetime.now())

    lines = filter(None, (line.strip() for line in proc.stdout.readlines()))
    all_output = console_to_str(b"\n".join(lines))
    if code:
        stderr_lines = filter(None, (line.strip() for line in proc.stderr.readlines()))
        all_output = console_to_str(b"".join(stderr_lines))
    output = "".join(all_output)
    if code != 0 and check_returncode:
        raise Exception(f"Command failed with code {code}: {cmd}")
    return output



"""Set up the ImJoy-Engine imjoy package."""
import os
from setuptools import setup, find_packages
import json

DESCRIPTION = (
    "ImJoy Plugin Engine for running Python plugins locally "
    "or remotely from ImJoy.io"
)


WORKER_REQUIREMENTS = [
    "python-engineio==3.9.1",
    "python-socketio[client]==4.4.0",
    "numpy",
    'janus;python_version>"3.3"',
    'pathlib;python_version<"3.4"',
]

REQUIREMENTS = WORKER_REQUIREMENTS + ["jupyter>=1.0.0"]

LEGACY_ENGINE_REQUIREMENTS = ["aiohttp", "aiohttp_cors", "gputil", "pyyaml"]

ROOT_DIR = os.path.dirname(__file__)
with open(os.path.join(ROOT_DIR, "README.md"), "r") as f:
    README = f.read()

with open(os.path.join(ROOT_DIR, "imjoy", "VERSION"), "r") as f:
    VERSION = json.load(f)["version"]

setup(
    name="imjoy",
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url="http://github.com/oeway/ImJoy-Engine",
    author="Wei OUYANG",
    author_email="oeway007@gmail.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require={
        "engine": LEGACY_ENGINE_REQUIREMENTS,
        "worker": [],
        "jupyter": [],
        "jupyter-worker": ["ipykernel>=5.1.4"],
    },
    zip_safe=False,
    entry_points={"console_scripts": ["imjoy = imjoy.__main__:main"]},
)



"""Provide utils for Python 3 plugins."""
import asyncio

from .utils import Promise


def make_coro(func):
    """Wrap a normal function with a coroutine."""

    async def wrapper(*args, **kwargs):
        """Run the normal function."""
        return func(*args, **kwargs)

    return wrapper


class FuturePromise(Promise, asyncio.Future):
    """Represent a promise as a future."""

    def __init__(self, pfunc, loop):
        """Set up promise."""
        self.loop = loop
        Promise.__init__(self, pfunc)
        asyncio.Future.__init__(self)

    def resolve(self, result):
        """Resolve promise."""
        if self._resolve_handler or self._finally_handler:
            super().resolve(result)
        else:
            self.loop.call_soon(self.set_result, result)

    def reject(self, error):
        """Reject promise."""
        if self._catch_handler or self._finally_handler:
            super().reject(error)
        else:
            if error:
                self.loop.call_soon(self.set_exception, Exception(str(error)))
            else:
                self.loop.call_soon(self.set_exception, Exception())
